from akasha.utils.atman import basic_llm
from akasha.utils.base import (
    DEFAULT_MODEL,
    DEFAULT_MAX_OUTPUT_TOKENS,
    DEFAULT_MAX_INPUT_TOKENS,
)
from akasha.utils.prompts.gen_prompt import (
    default_ask_prompt,
    default_conclusion_prompt,
    format_sys_prompt,
    format_image_prompt,
)
from akasha.utils.prompts.format import handle_params, handle_metrics, handle_table
from akasha.utils.db.load_docs import load_docs_from_info
from akasha.helper.base import get_doc_length
from akasha.helper.preprocess_prompts import merge_history_and_prompt
from akasha.helper.run_llm import (
    call_model,
    call_stream_model,
    call_batch_model,
    call_image_model,
    check_relevant_answer,
)
from typing import Union, List, Tuple, Generator
from pathlib import Path
from langchain.schema import Document
import time
import datetime
from akasha.helper.token_counter import myTokenizer


class ask(basic_llm):
    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        max_input_tokens: int = DEFAULT_MAX_INPUT_TOKENS,
        max_output_tokens: int = DEFAULT_MAX_OUTPUT_TOKENS,
        temperature: float = 0.0,
        prompt_format_type: str = "auto",
        language: str = "ch",
        record_exp: str = "",
        system_prompt: str = "",
        keep_logs: bool = False,
        verbose: bool = False,
        stream: bool = False,
        env_file: str = "",
    ):
        """_summary_

        Args:
            model (str, optional): _description_. Defaults to DEFAULT_MODEL.
            max_input_tokens (int, optional): _description_. Defaults to DEFAULT_MAX_INPUT_TOKENS.
            max_output_tokens (int, optional): _description_. Defaults to DEFAULT_MAX_OUTPUT_TOKENS.
            temperature (float, optional): _description_. Defaults to 0.0.
            language (str, optional): _description_. Defaults to "ch".
            record_exp (str, optional): _description_. Defaults to "".
            system_prompt (str, optional): _description_. Defaults to "".
            keep_logs (bool, optional): _description_. Defaults to False.
            verbose (bool, optional): _description_. Defaults to False.
            env_file (str, optional): _description_. Defaults to "".
        """
        super().__init__(
            model=model,
            max_input_tokens=max_input_tokens,
            max_output_tokens=max_output_tokens,
            temperature=temperature,
            language=language,
            record_exp=record_exp,
            system_prompt=system_prompt,
            keep_logs=keep_logs,
            verbose=verbose,
            env_file=env_file,
        )
        self.stream = stream
        self.prompt_format_type = prompt_format_type
        self.prompt = ""
        self.response = ""
        self.docs = []
        self.prompt_tokens, self.prompt_length = 0, 0
        self.doc_tokens, self.doc_length = 0, 0

        ## set default RAG prompt ##
        if self.system_prompt.replace(" ", "") == "":
            self.system_prompt = default_ask_prompt(self.language)

    def _display_info(self, batch: int = 1) -> bool:
        """display the information of the parameters if verbose is True"""
        if self.verbose is False:
            return False
        print(f"Model: {self.model}, Temperature: {self.temperature}")
        print(
            f"Prompt format type: {self.prompt_format_type}, Max input tokens: {self.max_input_tokens}"
        )
        print(
            f"Prompt tokens: {self.prompt_tokens}, Prompt length: {self.prompt_length}"
        )
        print(f"Doc tokens: {self.doc_tokens}, Doc length: {self.doc_length}")
        print(f"Batch:  {max(batch, 1)}\n\n")

        return True

    def _add_basic_log(
        self, timestamp: str, fn_type: str, history_messages: list = []
    ) -> bool:
        """add to logs for function if keep_logs is True"""
        if super()._add_basic_log(timestamp, fn_type) is False:
            return False

        self.logs[timestamp]["prompt"] = self.prompt
        self.logs[timestamp]["history_messages"] = history_messages
        return True

    def _add_result_log(self, timestamp: str, time: float) -> bool:
        """add to logs for ask function if keep_logs is True"""
        if super()._add_result_log(timestamp, time) is False:
            return False

        ### add token information ###
        self.logs[timestamp]["response"] = self.response
        self.logs[timestamp]["prompt_tokens"] = self.prompt_tokens
        self.logs[timestamp]["prompt_length"] = self.prompt_length
        self.logs[timestamp]["doc_tokens"] = self.doc_tokens
        self.logs[timestamp]["doc_length"] = self.doc_length

        return True

    def _add_result_log_vision(
        self, timestamp: str, time: float, image_path: str
    ) -> bool:
        """add to logs for vision function if keep_logs is True"""
        if super()._add_result_log(timestamp, time) is False:
            return False

        ### add token information ###
        self.logs[timestamp]["image_path"] = image_path
        self.logs[timestamp]["response"] = self.response
        self.logs[timestamp]["prompt_tokens"] = self.prompt_tokens
        self.logs[timestamp]["prompt_length"] = self.prompt_length

        return True

    def __call__(
        self,
        prompt: str,
        info: Union[str, list, Path, Document] = "",
        history_messages: List[str] = [],
        **kwargs,
    ) -> str:
        """the function to ask model with prompt and info documents,
        the info can be file path, url, directory path, or list of Document object
        if the info has too many tokens, it will be separated into multiple documents and ask model batchly

        Args:
            prompt (str): the user question
            info (Union[str, list], optional): the support information for llm to answer the question. Defaults to "".
            history_messages (list, optional): the chat history, record them and add to here for memory.
            Defaults to []. ex: ["hello! how are you?", "I am fine, thank you!"]

        Returns:
            str: the answer from llm
        """

        self._set_model(**kwargs)
        self._change_variables(**kwargs)
        self.prompt = prompt
        self.response = ""

        start_time = time.time()
        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        self.docs = load_docs_from_info(info)
        ### check if prompt <= max_input_tokens ###
        tot_prompts = self.prompt + self.system_prompt + "\n\n".join(history_messages)
        self.prompt_length = get_doc_length(self.language, tot_prompts)
        self.prompt_tokens = myTokenizer.compute_tokens(tot_prompts, self.model) + 10

        if self.prompt_tokens > self.max_input_tokens:
            print("\n\nThe tokens of prompt is larger than max_input_tokens.\n\n")
            raise ValueError("The tokens of prompt is larger than max_input_tokens.")

        self._add_basic_log(timestamp, "ask", history_messages=history_messages)

        ### separate documents and count tokens ###
        cur_documents, self.doc_tokens = self._separate_docs()
        self.doc_length = get_doc_length(self.language, "".join(cur_documents))

        prod_sys_prompts = self._process_batch_prompts(cur_documents, history_messages)

        self._display_info(
            len(cur_documents)
        )  # display the information of the parameters
        self._display_docs()
        ### start to ask llm ###
        if len(cur_documents) > 1:
            ## call batch model ##
            batch_responses = call_batch_model(
                self.model_obj,
                prod_sys_prompts,
            )
            fnl_conclusion_prompt = default_conclusion_prompt(prompt, self.language)
            ## check relevant answer if batch_responses > 10 ##
            if len(batch_responses) > 10:
                batch_responses = check_relevant_answer(
                    self.model_obj,
                    batch_responses,
                    self.prompt,
                    self.prompt_format_type,
                )

            batch_responses, cur_len = _retri_max_texts(
                batch_responses,
                self.max_input_tokens
                - myTokenizer.compute_tokens(fnl_conclusion_prompt, self.model),
                self.model,
            )
            fnl_input = format_sys_prompt(
                fnl_conclusion_prompt,
                "\n\n".join(batch_responses),
                self.prompt_format_type,
                self.model,
            )
            if self.stream:
                return self._display_stream(
                    fnl_input,
                )

            self.response = call_model(self.model_obj, fnl_input)

        else:
            if self.stream:
                return self._display_stream(prod_sys_prompts[0])

            self.response = call_model(self.model_obj, prod_sys_prompts[0])

        end_time = time.time()
        self._add_result_log(timestamp, end_time - start_time)
        self._upload_logs(end_time - start_time, self.doc_length, self.doc_tokens)
        return self.response

    def vision(self, prompt: str, image_path: Union[List[str], str], **kwargs) -> str:
        """ask model with image and prompt, image_path can be list of image path or url

        Args:
            image_path (Union[List[str], str]): image path or url (recommand jpeg or png file)
            prompt (str): user question

        Returns:
            str: _description_
        """
        self._set_model(**kwargs)
        self._change_variables(**kwargs)
        self.prompt = prompt
        fnl_input = []
        start_time = time.time()
        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        self._add_basic_log(timestamp, "vision")

        ## check model ##
        model_prefix = self.model.split(":")[0]
        if (
            model_prefix in ["hf", "hugginface", "anthropic", "claude", "anthro"]
            and self.stream is True
        ):
            raise ValueError(
                f"Currently {model_prefix} model does not support stream mode.\n\n"
            )

        if model_prefix in [
            "llama-cpu",
            "llama-gpu",
            "llama",
            "llama2",
            "llama-cpp",
            "chatglm",
            "chatglm2",
            "glm",
            "lora",
            "peft",
            "gptq",
            "gptq2",
        ]:
            raise ValueError(
                f"Currently {self.model} model does not support image input.\n\n"
            )

        ## count prompt tokens ##
        self.prompt_tokens = myTokenizer.compute_tokens(self.prompt, self.model)
        self.prompt_length = get_doc_length(self.language, self.prompt)

        ## decide prompt format ##
        if model_prefix in ["hf", "huggingface"]:
            fnl_input = format_image_prompt(image_path, prompt, "image_llama")

        elif model_prefix in ["anthropic", "claude", "anthro"]:
            fnl_input = format_image_prompt(image_path, prompt, "image_anthropic")

        elif model_prefix in ["gemini", "google"]:
            fnl_input = format_image_prompt(image_path, prompt, "image_gemini")
        else:
            fnl_input = format_image_prompt(image_path, prompt, "image_gpt")

        if self.stream:
            return self._display_stream(
                fnl_input,
            )
        self.response = call_image_model(self.model_obj, fnl_input)
        self._add_result_log_vision(timestamp, time.time() - start_time, image_path)

        return self.response

    def _display_stream(
        self, text_input: Union[str, List[str]]
    ) -> Generator[str, None, None]:
        ret = call_stream_model(
            self.model_obj,
            text_input,
        )

        for s in ret:
            self.response += s
            yield s

    def _separate_docs(
        self,
    ) -> Tuple[List[str], int]:
        """separate documents if the total length of documents exceed the max_input_tokens

        Returns:
            ret (List[str]): list of string of separated documents texts
            tot_len (int): the length of total documents

        """
        tot_token_len = 0
        cur_len = 0

        left_tokens = self.max_input_tokens - self.prompt_tokens
        ret = [""]
        for db_doc in self.docs:
            cur_token_len = myTokenizer.compute_tokens(db_doc.page_content, self.model)

            if cur_len + cur_token_len > left_tokens:
                if cur_token_len <= left_tokens:
                    cur_len = cur_token_len
                    ret.append(db_doc.page_content)
                else:
                    new_docs = self._truncate_docs(db_doc.page_content)
                    ret.extend(new_docs)
                    ret.append("")
                    cur_len = 0

                tot_token_len += cur_token_len
                continue

            cur_len += cur_token_len
            ret[-1] += db_doc.page_content + "\n"
            tot_token_len += cur_token_len

        ## remove the last empty string ##
        if ret[-1] == "":
            ret = ret[:-1]

        return ret, tot_token_len

    def _truncate_docs(self, text: str) -> List[str]:
        """truncate documents if the total length of documents exceed the max_input_tokens

        Returns:
            text (str): string of documents texts

        """

        new_docs = []
        tot_len = len(text)
        idx = 2
        truncate_content = text[: (tot_len // idx)]
        # truncate_len = helper.get_doc_length(self.language, truncate_content)
        truncated_token_len = myTokenizer.compute_tokens(truncate_content, self.model)
        while truncated_token_len > self.max_input_tokens:
            idx *= 2
            truncate_content = text[: (tot_len // idx)]
            truncated_token_len = myTokenizer.compute_tokens(
                truncate_content, self.model
            )

        rge = tot_len // idx
        st = 0
        ed = rge
        while st < tot_len:
            new_docs.append(text[st:ed])
            st += rge
            ed += rge

        return new_docs

    def _process_batch_prompts(
        self, cur_documents: List[str], history_messages: List[str] = []
    ) -> List[str]:
        """_summary_

        Args:
            prompts (List[str]): _description_

        Returns:
            List[str]: _description_
        """
        prod_sys_prompts = []

        if len(cur_documents) == 0:
            prod_sys_prompt = merge_history_and_prompt(
                history_messages,
                self.system_prompt,
                "User question: " + self.prompt,
                self.prompt_format_type,
                model=self.model,
            )
            prod_sys_prompts.append(prod_sys_prompt)
            return prod_sys_prompts

        for d_count in range(len(cur_documents)):
            prod_sys_prompt = merge_history_and_prompt(
                history_messages,
                self.system_prompt,
                "Reference document: "
                + cur_documents[d_count]
                + "\n\nUser question: "
                + self.prompt,
                self.prompt_format_type,
                model=self.model,
            )
            prod_sys_prompts.append(prod_sys_prompt)

        return prod_sys_prompts

    def _upload_logs(self, tot_time: float, doc_len: int, doc_tokens: int) -> str:
        """_summary_

        Args:
            tot_time (float): _description_
            doc_len (int): _description_
            doc_tokens (int): _description_
        """
        if self.record_exp == "":
            return "no record_exp assigned, so no logs uploaded"

        params = handle_params(
            self.model,
            self.language,
            "",
        )
        metrics = handle_metrics(doc_len, tot_time, doc_tokens)
        table = handle_table(self.prompt, self.docs, self.response)
        from akasha.utils.upload import aiido_upload

        aiido_upload(self.record_exp, params, metrics, table)

        return "logs uploaded"


def _retri_max_texts(
    texts_list: list, left_token_len: int, model_name: str = "openai:gpt-3.5-turbo"
) -> Tuple[list, int]:
    """return list of texts that do not exceed the left_token_len

    Args:
        texts_list (list): _description_
        left_doc_len (int): _description_

    Returns:
        Tuple[list, int]: _description_
    """
    ret = []
    cur_len = 0
    for text in texts_list:
        txt_len = myTokenizer.compute_tokens(text, model_name)
        if cur_len + txt_len > left_token_len:
            break
        cur_len += txt_len
        ret.append(text)
    return ret, cur_len



from typing import Callable, List, Union
from langchain_core.embeddings import Embeddings
from langchain_chroma import Chroma
from chromadb.config import Settings
from pathlib import Path
from akasha.utils.db.db_structure import dbs, get_storage_directory
from akasha.helper.handle_objects import handle_embeddings_and_name
from akasha.helper import separate_name, get_mac_address
import logging
from tqdm import tqdm
import datetime


def update_db(
    db: dbs,
    data_source: Union[List[Union[str, Path]], Union[Path, str]],
    embeddings: Union[str, Embeddings, Callable],
    chunk_size: int,
    env_file: str = "",
):
    embeddings, embeddings_name = handle_embeddings_and_name(
        embeddings, False, env_file
    )
    embed_type, embed_name = separate_name(embeddings_name)
    if not isinstance(data_source, list):
        data_source = [data_source]

    ## put db.id and db.metadatas into a dict
    db_dict = {}
    for i in range(len(db.metadatas)):
        db_dict[db.ids[i]] = db.metadatas[i]

    progress = tqdm(total=len(data_source), desc="upadte db")

    for data_path in data_source:
        db_path = get_storage_directory(data_path, chunk_size, embed_type, embed_name)

        if not Path(db_path).exists():
            logging.warning(f"db path {db_path} not found")
            print(f"db path {db_path} not found")
            continue

        progress.update(1)
        cur_meta = []
        client_settings = Settings(
            is_persistent=True,
            persist_directory=db_path,
            anonymized_telemetry=False,
        )
        docsearch = Chroma(persist_directory=db_path, client_settings=client_settings)
        doc_search_data = docsearch.get(include=["metadatas"])

        ### find the ids in db_dict that are in doc_search_data["ids"]
        for id in doc_search_data["ids"]:
            if id in db_dict:
                cur_meta.append(db_dict[id])
                del db_dict[id]

        ## update the db with the new metadata
        docsearch._collection.update(ids=doc_search_data["ids"], metadatas=cur_meta)
        docsearch._client._system.stop()
        docsearch = None
        del docsearch

    return


def add_chunks(
    data: Union[List[dict], dict],
    file_name: Union[str, Path],
    embeddings: Union[str, Embeddings, Callable],
    chunk_size: int,
    env_file: str = "",
):
    """add chunk(s) to the chromadb collection of the given file_name

    Args:
        data (Union[List[dict], dict]): data is list of dicts or a single dict with keys "text", "search_text", and "metadata". The "text" is the text to be added, "embedding" is the embedding vector, and "metadata" is a dictionary with additional information.
        file_name (Union[str, Path]): _description_
        embeddings (Union[str, Embeddings, Callable]): _description_
        chunk_size (int): _description_
    """

    if isinstance(data, dict):
        add_data = [data]
    else:
        add_data = data
    mac_address = get_mac_address()
    formatted_date = datetime.datetime.now().strftime("%Y-%m-%d-%H_%M_%S_%f")
    embeddings, embeddings_name = handle_embeddings_and_name(
        embeddings, False, env_file
    )
    embed_type, embed_name = separate_name(embeddings_name)
    db_path = get_storage_directory(file_name, chunk_size, embed_type, embed_name)
    client_settings = Settings(
        is_persistent=True,
        persist_directory=db_path,
        anonymized_telemetry=False,
    )
    docsearch = Chroma(persist_directory=db_path, client_settings=client_settings)
    metadatas = []
    texts = []
    vectors = []

    for d in add_data:
        metadatas.append(d.get("metadata", {}))

        if "text" not in d:
            print(f"text not found in data {d}, skipping")
            continue

        if "search_text" not in d:
            vectors.append(embeddings.embed_query(d["text"]))
        else:
            vectors.append(embeddings.embed_query(d["search_text"]))

        texts.append(d["text"])

    ids = [formatted_date + "_" + str(i) + "_" + mac_address for i in range(len(texts))]
    docsearch._collection.add(
        embeddings=vectors,
        metadatas=metadatas,
        documents=texts,
        ids=ids,
    )

    docsearch._client._system.stop()
    docsearch = None
    del docsearch

    return ids



from langchain.docstore.document import Document
from typing import Union
from pathlib import Path
import re


class dbs:
    def __init__(self, chrdb=[]):
        self.ids = []
        self.embeds = []
        self.metadatas = []
        self.docs = []
        self.vis = set()
        if isinstance(chrdb, list):
            pass

        else:
            data = chrdb.get(include=["embeddings", "metadatas", "documents"])
            if "ids" in data:
                self.ids = data["ids"]
                self.vis = set(data["ids"])
            if "embeddings" in data:
                self.embeds = data["embeddings"]
            else:
                self.embeds = [[] for _ in range(len(data["ids"]))]
            if "metadatas" in data:
                self.metadatas = data["metadatas"]
            else:
                self.metadatas = [{} for _ in range(len(data["ids"]))]
            if "documents" in data:
                self.docs = data["documents"]
            else:
                self.docs = ["" for _ in range(len(data["ids"]))]

    def merge(self, db: "dbs"):
        for i in range(len(db.ids)):
            if db.ids[i] not in self.vis:
                self.ids.append(db.ids[i])
                self.embeds.append(db.embeds[i])
                self.metadatas.append(db.metadatas[i])
                self.docs.append(db.docs[i])
                self.vis.add(db.ids[i])
        # self.ids.extend(db.ids)
        # self.embeds.extend(db.embeds)
        # self.metadatas.extend(db.metadatas)
        # self.docs.extend(db.docs)

    def add_chromadb(self, chrdb):
        data = chrdb.get(include=["embeddings", "metadatas", "documents"])
        if "ids" in data:
            self.ids.extend(data["ids"])

        if "embeddings" in data:
            self.embeds.extend(data["embeddings"])
        else:
            self.embeds.extend([[] for _ in range(len(data["ids"]))])
        if "metadatas" in data:
            self.metadatas.extend(data["metadatas"])
        else:
            self.metadatas.extend([{} for _ in range(len(data["ids"]))])
        if "documents" in data:
            self.docs.extend(data["documents"])
        else:
            self.docs.extend(["" for _ in range(len(data["ids"]))])

    def get_Documents(self):
        return [
            Document(page_content=self.docs[i], metadata=self.metadatas[i])
            for i in range(len(self.docs))
        ]

    def get_docs(self):
        return self.docs

    def get_ids(self):
        return self.ids

    def get_metadatas(self):
        return self.metadatas

    def get_embeds(self):
        return self.embeds


NO_PARENT_DIR_NAME = "NoPaReNtDiR"
FILE_LAST_CHANGE_FILE_NAME = "file_last_changed.json"
TEXT_EXTENSIONS = ["pdf", "md", "docx", "txt", "csv", "pptx"]
ALREADY_BUILT = "already_built"
NOT_BUILT = "not_built"
OLD_BUILT = "old_built"
HNSW_THRESHOLD = 400000


def get_storage_directory(
    dir_path: Union[Path, str],
    chunk_size: int,
    embed_type: str,
    embed_name: str,
) -> str:
    if isinstance(dir_path, str):
        if is_url(dir_path):
            sanitized_dir_path = _sanitize_path_string(dir_path)
            storage_directory = (
                "chromadb/"
                + sanitized_dir_path
                + "_"
                + embed_type
                + "_"
                + embed_name.replace("/", "-")
                + "_"
                + str(chunk_size)
            )
            return storage_directory
        else:
            dir_path = Path(dir_path)

    if dir_path != Path("."):
        db_dir = "-".join(
            [part.replace(" ", "").replace("_", "") for part in dir_path.parts if part]
        )
    else:
        db_dir = NO_PARENT_DIR_NAME

    storage_directory = (
        "chromadb/"
        + db_dir
        + "_"
        + embed_type
        + "_"
        + embed_name.replace("/", "-")
        + "_"
        + str(chunk_size)
    )

    return storage_directory


def _sanitize_path_string(path: str, max_len: int = 30) -> str:
    # Remove special characters and limit length to 30
    sanitized = re.sub(r"[^a-zA-Z0-9]", "", path)[:max_len]
    return sanitized


def is_url(pattern: str) -> bool:
    return bool(re.match(r"^(http|https)://", pattern))



# coding:utf-8
from rouge import Rouge
import rouge_chinese
import jieba
import warnings
from akasha.utils.prompts.gen_prompt import format_sys_prompt, format_llm_score
from akasha.utils.prompts.format import language_dict
from .handle_objects import handle_model_and_name
from .run_llm import call_model
import re
from langchain_core.language_models.base import BaseLanguageModel
from typing import Union
import importlib

warnings.filterwarnings("ignore")
jieba.setLogLevel(jieba.logging.INFO)  ## ignore logging jieba model information


def get_bert_pack():
    bert_score = importlib.import_module("bert_score")
    return bert_score.score


def get_bert_score(
    candidate_str: str, reference_str: str, language: str = "ch", round_digit: int = 3
):
    """bert score using pre-trained contextual embeddings from BERT to calculate the cosine similarity between two sentences.
    So different words with similar meaning will have higher score.

    Args:
        **candidate_str (str)**: the respones generated by llm you want to test the performance.\n
        **reference_str (str)**: the default answer string.\n
        **language (str, optional)**: texts language. Defaults to "ch".\n
        **round_digit (int, optional)**: round the score into which digit. Defaults to 3.\n

    Returns:
        float: bert score
    """

    score = get_bert_pack()
    try:
        if "chinese" in language_dict[language]:
            P, R, F1 = score([candidate_str], [reference_str], lang="zh", verbose=False)
        else:
            P, R, F1 = score([candidate_str], [reference_str], lang="en", verbose=False)
    except Exception:
        F1 = 0.0
    # round float into 3 digits behind 0
    F1 = round(float(F1), round_digit)

    return F1


def get_rouge_score(
    candidate_str: str, reference_str: str, language: str = "ch", round_digit: int = 3
):
    """use jieba to separate words from chinese sentence, and then use rouge_l to calculate the rouge score
    the difference between bleu and rouge is that bleu is focus on precision, but rouge is focus on the recall.

    Args:
        **candidate_str (str)**: the respones generated by llm you want to test the performance.\n
        **reference_str (str)**: the default answer string.\n
        **langugage (str, optional)**: texts language. Defaults to "ch".\n
        **round_digit (int, optional)**: round the score into which digit. Defaults to 3.\n

    Returns:
        float: rouge score
    """
    try:
        if "chinese" in language_dict[language]:
            rouge = rouge_chinese.Rouge(metrics=["rouge-l"])
            cand = " ".join(jieba.cut(candidate_str))
            ref = " ".join(jieba.cut(reference_str))
        else:
            rouge = Rouge(metrics=["rouge-l"])
            cand = candidate_str
            ref = reference_str

        F1 = rouge.get_scores(cand, ref)[0]["rouge-l"]["f"]
    except Exception:
        F1 = 0.0
    F1 = round(F1, round_digit)

    return F1


def get_llm_score(
    candidate_str: str,
    reference_str: str,
    model: Union[BaseLanguageModel, str],
    prompt_format_type: str = "auto",
    round_digit: int = 3,
):
    """use LLM to calculate the score of the candidate string and reference string.

    Args:
        candidate_str (str): _description_
        reference_str (str): _description_
        model (Union[BaseLanguageModel, str]): _description_
        prompt_format_type (str, optional): _description_. Defaults to "auto".
        round_digit (int, optional): _description_. Defaults to 3.

    Returns:
        _type_: _description_
    """

    model, model_name = handle_model_and_name(model)
    system_prompt, prompt = format_llm_score(candidate_str, reference_str)
    input_text = format_sys_prompt(
        system_prompt, prompt, prompt_format_type, model_name
    )

    response = call_model(model, input_text)

    # find the first float number in the response string and turn to float
    try:
        score = round(float(re.findall(r"\d+\.?\d*", response)[0]), round_digit)
    except Exception:
        score = 0.0
    return score


def get_toxic_score(texts: str, round_digit: int = 3):
    """check the toxic score of the input texts.

    Args:
        texts (str): _description_
        round_digit (int, optional): _description_. Defaults to 3.

    Returns:
        _type_: _description_
    """
    from transformers import pipeline

    pipe = pipeline("text-classification", model="martin-ha/toxic-comment-model")
    res = pipe.predict(texts)[0]

    if res["label"] == "toxic":
        score = round(float(res["score"]), round_digit)
    else:
        score = round(1 - float(res["score"]), round_digit)

    return score



from typing import List, Any, Optional
from langchain.llms.base import LLM
from transformers import AutoTokenizer, AutoModel

from pydantic import Field


class chatGLM(LLM):
    max_token: int = 4096
    max_output_tokens: int = 1024
    temperature: float = 0.01
    top_p: float = 0.95
    history: list = []
    tokenizer: Any = Field(default=None)
    model: Any = Field(default=None)
    model_name: str = ""

    def __init__(
        self, model_name: str, temperature: float = 0.01, max_output_tokens: int = 1024
    ):
        """define chatglm model and the tokenizer

        Args:
            **model_name (str)**: chatglm model name\n
        """
        if model_name == "":
            model_name = "THUDM/chatglm2-6b"

        super().__init__()
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name, trust_remote_code=True
        )
        self.model = AutoModel.from_pretrained(
            model_name, trust_remote_code=True, device="cuda"
        )
        self.max_output_tokens = max_output_tokens
        self.temperature = temperature
        if self.temperature == 0.0:
            self.temperature = 0.01
        self.model_name = model_name

    @property
    def _llm_type(self) -> str:
        """return llm type

        Returns:
            str: llm type
        """
        return f"chatglm:{self.model_name}  ChatGLM"

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        self.model = self.model.eval()
        response, history = self.model.chat(self.tokenizer, prompt, history=[])
        return response



from typing import Union
from akasha.utils.prompts.gen_prompt import (
    format_sys_prompt,
    format_history_prompt,
    decide_auto_prompt_format_type,
)
from typing import Tuple
from .token_counter import myTokenizer


def merge_history_and_prompt(
    history_messages: list,
    system_prompt: str,
    prompt: str,
    prompt_format_type: str = "auto",
    user_tag: str = "user",
    assistant_tag: str = "assistant",
    model: str = "remote:xxx",
) -> Union[str, list]:
    """merge system prompt, history messages, and prompt based on the prompt format type, if history_messages is empty, return the prompt
    if prompt_format_type is start with "chat_", it will become list of dictionary, otherwise it will become string

    Args:
        history_messages (list): _description_
        system_prompt (str): _description_
        prompt (str): _description_
        prompt_format_type (str, optional): _description_. Defaults to "gpt".
        user_tag (str, optional): _description_. Defaults to "user".
        assistant_tag (str, optional): _description_. Defaults to "assistant".
        model (_type_, optional): _description_. Defaults to "remote:xxx".

    Returns:
        Union[str, list]: _description_
    """
    ### decide prompt format type if auto###
    if prompt_format_type == "auto":
        prompt_format_type = decide_auto_prompt_format_type(model)

    ### if history_messages is empty, return the prompt ###
    if history_messages == [] or history_messages is None or history_messages == "":
        return format_sys_prompt(system_prompt, prompt, prompt_format_type)

    if "chat_" in prompt_format_type and prompt_format_type != "chat_gemma":
        if prompt_format_type == "chat_gemini":
            assistant_tag = "model"

        text_input = format_sys_prompt(system_prompt, "", prompt_format_type)

        prod_prompt = format_sys_prompt("", prompt, prompt_format_type)

        prod_history = format_history_prompt(
            history_messages, prompt_format_type, user_tag, assistant_tag
        )

        text_input.extend(prod_history)

        text_input.extend(prod_prompt)

        return text_input

    else:
        history_str = ""

        for i in range(len(history_messages)):
            if i % 2 == 0:
                history_str += user_tag + ": " + history_messages[i] + "\n"

            else:
                history_str += assistant_tag + ": " + history_messages[i] + "\n"

        history_str += "\n\n"
        return format_sys_prompt(
            system_prompt, history_str + prompt, prompt_format_type
        )


def retri_history_messages(
    messages: list,
    pairs: int = 10,
    max_input_tokens: int = 1500,
    model_name: str = "openai:gpt-3.5-turbo",
    role1: str = "User",
    role2: str = "Assistant",
) -> Tuple[str, int]:
    """from messages dict list, get pairs of user question and assistant response from most recent and not exceed max_input_tokens and pairs, and return the text with total length

    Args:
        messages (list): history messages list, each index is a dict with keys: role("user", "assistant"), content(content of message)
        pairs (int, optional): the maximum number of messages. Defaults to 10.
        max_input_tokens (int, optional): the maximum number of messages tokens. Defaults to 1500.
        model_name (str, optional): model name. Defaults to "openai:gpt-3.5-turbo".
        role1 (str, optional): role1 name. Defaults to "User".
        role2 (str, optional): role2 name. Defaults to "Assistant".

    Returns:
        Tuple[str, int]: return the text with total length.
    """
    cur_len = 0
    count = 0
    ret = []
    splitter = "\n----------------\n"

    for i in range(len(messages) - 1, -1, -2):
        if count >= pairs:
            break
        if (messages[i]["role"] != role2) or (messages[i - 1]["role"] != role1):
            i += 1
            continue
        texta = f"{role2}: " + messages[i]["content"].replace("\n", "") + "\n"
        textq = f"{role1}: " + messages[i - 1]["content"].replace(
            "\n", ""
        )  # {(i+1)//2}.
        len_texta = myTokenizer.compute_tokens(texta, model_name)
        len_textq = myTokenizer.compute_tokens(textq, model_name)

        if cur_len + len_texta > max_input_tokens:
            break
        cur_len += len_texta
        ret.append(texta)

        if cur_len + len_textq > max_input_tokens:
            break
        cur_len += len_textq
        ret.append(textq)

        count += 1

    if count == 0:
        return "", 0

    ret.reverse()
    ret_str = splitter + "".join(ret) + splitter

    return ret_str, cur_len



from typing import Any, List, Optional, Tuple
from pydantic import Field
import numpy as np
from langchain.embeddings.base import Embeddings
from langchain.schema import BaseRetriever, Document
from akasha.utils.db.db_structure import dbs
from langchain_community.utils.math import cosine_similarity


class myMMRRetriever(BaseRetriever):
    embeddings: Embeddings = Field(default=None)
    """Embeddings model to use."""
    index: Any = Field(default=None)
    """Index of embeddings."""
    texts: List[str] = Field(default=None)
    """List of texts to index."""
    metadata: List[dict] = Field(default=None)
    k: int = 3
    """Number of results to return."""
    relevancy_threshold: Optional[float] = None
    log: dict = {}
    lambda_mult: float = 0.5

    @classmethod
    def from_db(
        cls,
        db: dbs,
        embeddings: Embeddings,
        k: int = 3,
        relevancy_threshold: float = 0.0,
        log: dict = {},
        lambda_mult: float = 0.5,
    ):
        # db_data = _get_all_docs(db)
        index = np.array(db.get_embeds())
        texts = db.get_docs()
        metadata = db.get_metadatas()
        return cls(
            embeddings=embeddings,
            index=index,
            texts=texts,
            metadata=metadata,
            k=k,
            relevancy_threshold=relevancy_threshold,
            log=log,
            lambda_mult=lambda_mult,
        )

    def _get_relevant_documents(self, query: str) -> List[Document]:
        return self._gs(query)[0]

    def _gs(self, query: str) -> Tuple[List[Document], List[float]]:
        """implement using custom function to find relevant documents, the custom function func should
        have four input.
            1. a np.array of embedding vectors of query query_embeds np.array)
            2. a np.array of np.array contain embedding vectors of query and documents docs_embeds (np.array)
            3. the number of topK return documents k (int)
            4. relevant threshold from 0.0 ~ 1.0 threshold (float)
        And the function func should return a list of index which length is equals to k, represent
        the index of documents that are most relevant to the input query.

        Args:
            **query (str)**: query string that used to find relevant documents\n

        Returns:
            List[Document]: relevant documents
        """

        top_k_results = []
        query_embeds = np.array(self.embeddings.embed_query(query))
        docs_embeds = self.index

        if min(self.k, len(docs_embeds)) <= 0:
            return [], []
        if query_embeds.ndim == 1:
            query_embeds = np.expand_dims(query_embeds, axis=0)
        similarity_to_query = cosine_similarity(query_embeds, docs_embeds)[0]
        most_similar = int(np.argmax(similarity_to_query))
        mmr_scores = [similarity_to_query[most_similar]]
        relevant_docs_idx = [most_similar]
        selected = np.array([docs_embeds[most_similar]])
        while len(relevant_docs_idx) < min(self.k, len(docs_embeds)):
            best_score = -np.inf
            idx_to_add = -1
            similarity_to_selected = cosine_similarity(docs_embeds, selected)
            for i, query_score in enumerate(similarity_to_query):
                if i in relevant_docs_idx:
                    continue
                redundant_score = max(similarity_to_selected[i])
                equation_score = (
                    self.lambda_mult * query_score
                    - (1 - self.lambda_mult) * redundant_score
                )
                if equation_score > best_score:
                    best_score = equation_score
                    idx_to_add = i
            relevant_docs_idx.append(idx_to_add)
            mmr_scores.append(best_score)
            selected = np.append(selected, [docs_embeds[idx_to_add]], axis=0)

        ### from index rebuild the documents ###
        for idx in relevant_docs_idx[: self.k]:
            top_k_results.append(
                Document(page_content=self.texts[idx], metadata=self.metadata[idx])
            )

        return top_k_results, mmr_scores[: self.k]

    async def _aget_relevant_documents(self, query: str) -> List[Document]:
        return self._gs(query)[0]

    def get_relevant_documents_and_scores(
        self,
        query: str,
    ) -> Tuple[List[Document], List[float]]:
        return self._gs(query)



from typing import Optional, Union, Dict, Sequence
import os
from pathlib import Path
from datetime import datetime
import pkg_resources


def aiido_upload(
    exp_name,
    params: dict = {},
    metrics: dict = {},
    table: dict = {},
    path_name: str = "",
):
    """upload params_metrics, table to mlflow server for tracking.

    Args:
        **exp_name (str)**: experiment name on the tracking server, if not found, will create one .\n
        **params (dict, optional)**: parameters dictionary. Defaults to {}.\n
        **metrics (dict, optional)**: metrics dictionary. Defaults to {}.\n
        **table (dict, optional)**: table dictionary, used to compare text context between different runs in the experiment. Defaults to {}.\n
    """
    try:
        import mlflow

    except ImportError as e:
        print("mlflow is not installed. Please install it with 'pip install mlflow'.")
        raise e

    time_now = datetime.now().strftime("%Y%m%d%H%M%S")

    if path_name is None:
        path_name = ""
    if "model" not in params or "embeddings" not in params:
        mlflow_init(experiment=exp_name, run=path_name + "-" + time_now)

    else:
        mod = params["model"].split(":")
        emb = params["embeddings"].split(":")[0]
        sea = params["search_type"]
        mlflow_init(
            experiment=exp_name,
            run=emb + "-" + sea + "-" + "-".join(mod) + "-" + time_now,
        )

    log_params_and_metrics(params=params, metrics=metrics)

    if len(table) > 0:
        mlflow.log_table(table, "table.json")
    mlflow.end_run()


def mlflow_init(
    experiment: Optional[str] = None,
    run: Optional[str] = None,
    load_config_from_dotenv: Union[bool, str] = False,
    do_not_raise: bool = True,
):
    try:
        import mlflow

        try:
            __version__ = pkg_resources.require("akasha-terminal")[0].version
        except Exception:
            __version__ = "dev"

    except ImportError:
        print("package is not installed. Please install it with 'pip install mlflow'.")
        if not do_not_raise:
            raise
        return

    try:
        if load_config_from_dotenv:
            from dotenv import load_dotenv

            if isinstance(load_config_from_dotenv, str):
                res = load_dotenv(load_config_from_dotenv, override=True)
            elif load_config_from_dotenv is True:
                res = load_dotenv(os.path.join(os.getcwd(), ".env"), override=True)
            else:
                res = False
            if not res:
                raise RuntimeError("No environment variables were loaded.")
    except Exception as e:
        print(
            f"Error loading environment variables from {load_config_from_dotenv if isinstance(load_config_from_dotenv, str) else '.env'}\n{e}"
        )

    try:
        time_now = datetime.now().strftime("%Y%m%d%H%M%S")
        mlflow.set_tracking_uri(
            os.getenv("TRACKING_SERVER_URI", Path(os.path.join(os.getcwd(), "mlruns")))
        )
        tracking_uri = mlflow.get_tracking_uri()
        if tracking_uri.endswith("mlruns") and tracking_uri.startswith("file:"):
            print(
                f"Remote TRACKING_SERVER_URI is not set, use default local value: {tracking_uri}"
            )

        if experiment is not None:
            mlflow.set_experiment(experiment)

        if run is None or run == "":
            run = "mlflow-" + time_now
        mlflow.set_tag("mlflow.runName", run)
        mlflow.set_tag("_Software.akasha.SDK", __version__)
        print(
            f"\n[{time_now}][OK][init] mlflow initialized with experiment: {experiment}, run: {run}"
        )
    except Exception as e:
        print(
            f"[{time_now}][Fail][init] Failed initializing mlflow with experiment: {experiment}, run: {run}"
        )
        if not do_not_raise:
            raise e


def log_params_and_metrics(params: Dict, metrics: Union[Dict, Sequence[Dict]]):
    """Equivalent to mlflow.log_params then mlflow.log_metrics

    Parameters
    ----------
    params : Dict
        Training parameters, such as learning rate, weight decay, etc.
    metrics : Dict
        Evaluation metrics, such as accuracy, top-5 accuracy, etc. or a list of such metrics
    """
    # Check if mlflow is installed
    import mlflow

    try:
        time_now = datetime.now().strftime("%Y%m%d%H%M%S")
        mlflow.log_params(params)

        try:
            if isinstance(metrics, Sequence):
                for metric in metrics:
                    mlflow.log_metrics(metric)
            else:
                mlflow.log_metrics(metrics)
            print(f"[{time_now}][OK][log_metrics] Metrics logged")
        except Exception as e:
            print(f"[{time_now}][Fail][log_metrics] {e}")
            raise e

        print(f"[{time_now}][OK][log_params_and_metrics] Params and metrics logged\n")

    except Exception as e:
        print(
            f"[{time_now}][Fail][log_params_and_metrics] Failed logging params and metrics: {e}\n"
        )
        raise e



from .retrievers.retri_rerank import rerank_reduce


def get_relevant_doc_auto(
    retriver_list: list,
    query: str,
) -> list:
    """try every solution to get  to search relevant documents.

    Args:
        **db (Chromadb)**: chroma db\n
        **query (str)**: the query str used to search similar documents\n
        **k (int)**: for each search type, return first k documents\n
        **times (int)**: the magnification of max embeddings(?)\n

    Returns:
        list: list of selected relevant Documents
    """

    # mmrR = retriver_list[0]
    # docs_mmr, mmr_scores = mmrR._gs(query)
    # print("MMR: ", mmr_scores, len(mmr_scores), "\n\n")
    # print(docs_mmr[:4])

    ### svm ###
    # svmR = retriver_list[0]
    # docs_svm, svm_scores = svmR._gs(query)
    # print("SVM: ", svm_scores, docs_svm[0], "\n\n")

    # ### tfidf ###

    # tfretriever = retriver_list[1]
    # docs_tf, tf_scores = tfretriever._gs(query)
    # print("TFIDF", tf_scores, docs_tf[0], "\n\n")

    # ### knn ###
    knnR = retriver_list[0]
    docs_knn, knn_scores = knnR._gs(query)
    # print("KNN: ", knn_scores, len(knn_scores), "\n\n")

    ### bm25 ###
    bm25R = retriver_list[1]
    docs_bm25, bm25_scores = bm25R._gs(query)
    # print("BM25: ", bm25_scores[:10], len(bm25_scores), "\n\n")

    ### decide which to use ###
    backup_docs = []
    final_docs = []  # docs_mmr[0]
    del bm25R
    ## backup_docs is all documents from docs_svm that svm_scores>0.2 ##
    low = 0
    for i in range(len(knn_scores)):
        if knn_scores[i] >= 0.95:
            backup_docs.append(docs_knn[i])
        else:
            low = i
            break

    if bm25_scores[0] >= 70:
        ## find out the idx that the sorted tf_scores is not 0
        idx = 0
        for i in range(len(bm25_scores)):
            if bm25_scores[i] < 70 or i >= 2:
                idx = i
                break
        final_docs.extend(docs_bm25[:idx])

    final_docs.extend(backup_docs)
    final_docs.extend(docs_knn[low:])
    return final_docs


def get_relevant_doc_auto_rerank(
    retriver_list: list,
    query: str,
    k: int,
) -> list:
    """try every solution to get  to search relevant documents.

    Args:
        **db (Chromadb)**: chroma db\n
        **query (str)**: the query str used to search similar documents\n
        **k (int)**: for each search type, return first k documents\n
        **times (int)**: the magnification of max embeddings(?)\n

    Returns:
        list: list of selected relevant Documents
    """

    ### knn ###
    knnR = retriver_list[0]
    docs_knn, knn_scores = knnR._gs(query)
    pr = max(int(0.1 * len(docs_knn)), 10)
    # print("KNN: ", knn_scores, docs_knn[0], "\n\n")

    ### bm25 ###
    bm25R = retriver_list[1]
    docs_bm25, bm25_scores = bm25R._gs(query)
    # print("BM25: ", bm25_scores[:10], len(bm25_scores), "\n\n")

    ### decide which to use ###
    backup_docs = []
    final_docs = []  # docs_mmr[0]
    del knnR, bm25R
    ## backup_docs is all documents from docs_svm that svm_scores>0.2 ##

    for i in range(len(knn_scores)):
        if knn_scores[i] >= 0.9:
            backup_docs.append(docs_knn[i])
        else:
            break

    if bm25_scores[0] >= 70:
        # if verbose:
        #     print("<<search>>go to bm25\n\n")

        ## find out the idx that the sorted tf_scores is not 0
        idx = 0
        for i in range(len(bm25_scores)):
            if bm25_scores[i] < 70 or i >= 2:
                idx = i
                break
        final_docs.extend(docs_bm25[:idx])

    if len(backup_docs) < pr:
        # if verbose:
        #     print("<<search>>go to knn\n\n")

        final_docs.extend(backup_docs)

    else:
        # if verbose:
        #     print("<<search>>go to knn+rerank\n\n")
        final_docs.extend(rerank_reduce(query, backup_docs, k))

    return final_docs



from fastapi import FastAPI
from typing import Optional, List, Dict, Any, Union
from pydantic import BaseModel
import akasha
import gc
import torch
import os

app = FastAPI()


def clean():
    try:
        gc.collect()
        torch.cuda.ipc_collect()
        torch.cuda.empty_cache()
    except Exception:
        pass


class InfoModel(BaseModel):
    prompt: str
    info: Union[str, list] = ""
    model: Optional[str] = "openai:gpt-3.5-turbo"
    system_prompt: Optional[str] = ""
    max_input_tokens: Optional[int] = 3000
    temperature: Optional[float] = 0.0
    env_config: Optional[Dict[str, Any]] = {}


class webInfoModel(InfoModel):
    search_engine: str = "wiki"
    search_num: int = 5


class ConsultModel(BaseModel):
    data_source: Union[str, List[str]]
    prompt: Union[str, List[Any]]
    chunk_size: Optional[int] = 1000
    model: Optional[str] = "openai:gpt-3.5-turbo"
    embedding_model: Optional[str] = "openai:text-embedding-ada-002"
    threshold: Optional[float] = 0.0
    search_type: Optional[str] = "auto"
    system_prompt: Optional[str] = ""
    max_input_tokens: Optional[int] = 3000
    temperature: Optional[float] = 0.0
    env_config: Optional[Dict[str, Any]] = {}


class ConsultModelReturn(BaseModel):
    response: Union[str, List[str]]
    status: str
    logs: Dict[str, Any]
    timestamp: str
    warnings: List[str] = []


class SummaryModel(BaseModel):
    content: Union[str, list] = ""
    summary_type: Optional[str] = "map_reduce"
    summary_len: Optional[int] = 500
    max_input_tokens: Optional[int] = 3200
    model: Optional[str] = "openai:gpt-3.5-turbo"
    system_prompt: Optional[str] = ""
    env_config: Optional[Dict[str, Any]] = {}


OTHER_ENV = set(
    [
        "SERPER_API_KEY",
        "BRAVE_API_KEY",
        "ANTHROPIC_API_KEY",
        "GEMINI_API_KEY",
    ]
)


def load_env(config: dict) -> bool:
    """delete old environment variable and load new one.

    Args:
        config (dict): dictionary may contain openai_key, azure_key, azure_base.

    Returns:
        bool: load success or not
    """

    for key, val in config.items():
        if key in OTHER_ENV:
            os.environ[key] = val

    if "OPENAI_API_KEY" in os.environ:
        del os.environ["OPENAI_API_KEY"]
    if "AZURE_API_BASE" in os.environ:
        del os.environ["AZURE_API_BASE"]
    if "AZURE_API_KEY" in os.environ:
        del os.environ["AZURE_API_KEY"]
    if "AZURE_API_TYPE" in os.environ:
        del os.environ["AZURE_API_TYPE"]
    if "AZURE_API_VERSION" in os.environ:
        del os.environ["AZURE_API_VERSION"]

    if "OPENAI_API_KEY" in config and config["OPENAI_API_KEY"] != "":
        os.environ["OPENAI_API_KEY"] = config["OPENAI_API_KEY"]

        return True

    if (
        "AZURE_API_KEY" in config
        and "AZURE_API_BASE" in config
        and config["AZURE_API_KEY"] != ""
        and config["AZURE_API_BASE"] != ""
    ):
        os.environ["AZURE_API_KEY"] = config["AZURE_API_KEY"]
        os.environ["AZURE_API_BASE"] = config["AZURE_API_BASE"]
        os.environ["AZURE_API_TYPE"] = "azure"
        os.environ["AZURE_API_VERSION"] = "2023-05-15"
        return True

    if (
        os.environ.get("ANTHROPIC_API_KEY", "") != ""
        or os.environ.get("GEMINI_API_KEY", "") != ""
    ):
        return True
    return False


@app.post("/RAG")
def RAG(user_input: ConsultModel):
    """run RAG for document search, load openai config if needed.

    Args:
        user_input (ConsultModel): data input class used in get_response
        data_source: Union[str, List[str]]
        prompt: Union[str, List[str]]
        chunk_size:Optional[int]=1000
        model:Optional[str] = "openai:gpt-3.5-turbo"
        threshold:Optional[float] (deprecated) = 0.0
        search_type:Optional[str] = 'auto'
        system_prompt:Optional[str] = ""
        temperature:Optional[float]=0.0
        env_config:Optional[Dict[str, Any]] = {}
        max_input_tokens: Optional[int] = 3000


    Returns:
        dict: status, response, logs
    """
    if user_input.model.split(":")[0] in [
        "openai",
        "gemini",
        "anthropic",
    ] or user_input.embedding_model.split(":")[0] in ["openai", "gemini", "anthropic"]:
        if not load_env(config=user_input.env_config):
            return {"status": "fail", "response": "load env config failed.\n\n"}
    try:
        clean()
        qa = akasha.RAG(
            verbose=True,
            search_type=user_input.search_type,
            model=user_input.model,
            temperature=user_input.temperature,
            max_input_tokens=user_input.max_input_tokens,
            embeddings=user_input.embedding_model,
            chunk_size=user_input.chunk_size,
            system_prompt=user_input.system_prompt,
        )

        response = qa(
            data_source=user_input.data_source, prompt=user_input.prompt, keep_logs=True
        )
    except Exception as e:
        err_message = e.__str__()
        response = None

    try:
        ig_files = qa.ignored_files
    except Exception:
        ig_files = []

    ## get logs
    timesp = ""
    if len(qa.timestamp_list) == 0:
        logs = {}
    else:
        timesp = qa.timestamp_list[-1]
        if timesp in qa.logs:
            logs = qa.logs[timesp]

    if response is None:
        user_output = ConsultModelReturn(
            response=f"text generation encounter errors, {err_message}\n",
            status="fail",
            logs=logs,
            timestamp=timesp,
            warnings=ig_files,
        )
    else:
        user_output = ConsultModelReturn(
            response=response,
            status="success",
            logs=logs,
            timestamp=timesp,
            warnings=ig_files,
        )

    del qa.model_obj
    del qa
    clean()
    return user_output


@app.post("/ask")
def ask(user_input: InfoModel):
    """run ask to ask llm, load openai config if needed.

    Args:
        user_input (InfoModel): data input class used in get_response
        info: Union[str, List[str]]
        prompt: str
        model:Optional[str] = "openai:gpt-3.5-turbo"
        system_prompt:Optional[str] = ""
        max_input_tokens:Optional[int]=3000
        temperature:Optional[float]=0.0
        env_config:Optional[Dict[str, Any]] = {}


    Returns:
        dict: status, response, logs
    """
    if user_input.model.split(":")[0] in ["openai", "gemini", "anthropic"]:
        if not load_env(config=user_input.env_config):
            return {"status": "fail", "response": "load env config failed.\n\n"}
    try:
        clean()
        qa = akasha.ask(
            verbose=True,
            model=user_input.model,
            temperature=user_input.temperature,
            max_input_tokens=user_input.max_input_tokens,
            system_prompt=user_input.system_prompt,
        )

        response = qa(prompt=user_input.prompt, info=user_input.info, keep_logs=True)

    except Exception as e:
        err_message = e.__str__()
        response = None

    ig_files = []

    ## get logs
    timesp = ""
    if len(qa.timestamp_list) == 0:
        logs = {}
    else:
        timesp = qa.timestamp_list[-1]
        if timesp in qa.logs:
            logs = qa.logs[timesp]

    if response is None:
        user_output = ConsultModelReturn(
            response=f"text generation encounter errors, {err_message}\n",
            status="fail",
            logs=logs,
            timestamp=timesp,
            warnings=ig_files,
        )
    else:
        user_output = ConsultModelReturn(
            response=response,
            status="success",
            logs=logs,
            timestamp=timesp,
            warnings=ig_files,
        )

    del qa.model_obj
    del qa
    clean()
    return user_output


@app.post("/summary")
def summary(user_input: SummaryModel):
    """get summary to summarize articles or files, load openai config if needed.

    Args:
        content(Union[str,List[str]]): file paths or urls, or text content
        summary_type: Optional[str] = "map_reduce" : summary method, "map_reduce" or "refine".
        summary_len: Optional[int] = 500 : the length of summary result
        model: Optional[str] = "openai:gpt-3.5-turbo"
        system_prompt: Optional[str] = ""
        env_config: Optional[Dict[str, Any]] = {}

    Returns:
        _type_: _description_
    """

    if user_input.model.split(":")[0] in ["openai", "gemini", "anthropic"]:
        if not load_env(config=user_input.env_config):
            return {"status": "fail", "response": "load env config failed.\n\n"}
    try:
        clean()
        sum = akasha.summary(
            chunk_size=500,
            chunk_overlap=50,
            model=user_input.model,
            verbose=True,
            system_prompt=user_input.system_prompt,
            max_input_tokens=user_input.max_input_tokens,
            sum_type=user_input.summary_type,
            sum_len=user_input.summary_len,
            temperature=0.0,
        )

        response = sum(content=user_input.content, keep_logs=True)
    except Exception as e:
        err_message = e.__str__()
        response = None

    ## get logs
    timesp = ""
    if len(sum.timestamp_list) == 0:
        logs = {}
    else:
        timesp = sum.timestamp_list[-1]
        if timesp in sum.logs:
            logs = sum.logs[timesp]

    if response is None:
        user_output = ConsultModelReturn(
            response=f"text generation encounter errors, {err_message}\n",
            status="fail",
            logs=logs,
            timestamp=timesp,
        )
    else:
        user_output = ConsultModelReturn(
            response=response, status="success", logs=logs, timestamp=timesp
        )
    del sum.model_obj
    del sum
    clean()

    return user_output


@app.post("/websearch")
def websearch(user_input: webInfoModel):
    """run websearch to use prompt search in web and answer the user question, load openai config if needed.

    Args:
        user_input (InfoModel): data input class used in get_response
        info: Union[str, List[str]]
        prompt: str
        model:Optional[str] = "openai:gpt-3.5-turbo"
        system_prompt:Optional[str] = ""
        max_input_tokens:Optional[int]=3000
        temperature:Optional[float]=0.0
        search_engine: str = "wiki"
        search_num: int = 5
        env_config:Optional[Dict[str, Any]] = {}


    Returns:
        dict: status, response, logs
    """

    load_env(config=user_input.env_config)

    try:
        clean()
        qa = akasha.websearch(
            verbose=True,
            model=user_input.model,
            temperature=user_input.temperature,
            max_input_tokens=user_input.max_input_tokens,
            system_prompt=user_input.system_prompt,
            search_engine=user_input.search_engine,
            search_num=user_input.search_num,
        )

        response = qa(prompt=user_input.prompt, keep_logs=True)

    except Exception as e:
        err_message = e.__str__()
        response = None

    ig_files = []

    ## get logs
    timesp = ""
    if len(qa.timestamp_list) == 0:
        logs = {}
    else:
        timesp = qa.timestamp_list[-1]
        if timesp in qa.logs:
            logs = qa.logs[timesp]

    if response is None:
        user_output = ConsultModelReturn(
            response=f"text generation encounter errors, {err_message}\n",
            status="fail",
            logs=logs,
            timestamp=timesp,
            warnings=ig_files,
        )
    else:
        user_output = ConsultModelReturn(
            response=response,
            status="success",
            logs=logs,
            timestamp=timesp,
            warnings=ig_files,
        )

    del qa.model_obj
    del qa
    clean()
    return user_output



from langchain_community.retrievers import SVMRetriever
from typing import Any, List, Optional, Tuple
from pydantic import Field
import numpy as np
from langchain.embeddings.base import Embeddings
from langchain.schema import BaseRetriever, Document
from akasha.utils.db.db_structure import dbs


class mySVMRetriever(BaseRetriever):
    embeddings: Embeddings = Field(default=None)
    """Embeddings model to use."""
    index: Any = Field(default=None)
    """Index of embeddings."""
    texts: List[str] = Field(default=None)
    """List of texts to index."""
    metadata: List[dict]
    k: int = 3
    """Number of results to return."""
    relevancy_threshold: Optional[float] = None

    @classmethod
    def from_db(
        cls,
        db: dbs,
        embeddings: Embeddings,
        k: int = 3,
        relevancy_threshold: float = 0.0,
        **kwargs: Any,
    ) -> SVMRetriever:
        index = np.array(db.get_embeds())
        texts = db.get_docs()
        metadata = db.get_metadatas()
        return cls(
            embeddings=embeddings,
            index=index,
            texts=texts,
            metadata=metadata,
            k=k,
            relevancy_threshold=relevancy_threshold,
            **kwargs,
        )

    def _gs(self, query: str) -> Tuple[List[Document], List[float]]:
        """implement svm to find relevant documents

        Args:
            **query (str)**: query string that used to find relevant documents\n

        Returns:
            List[Document]: relevant documents
        """
        try:
            from sklearn import svm
        except ImportError:
            raise ImportError(
                "Could not import scikit-learn, please install with `pip install "
                "scikit-learn`."
            )

        query_embeds = np.array(self.embeddings.embed_query(query))
        x = np.concatenate([query_embeds[None, ...], self.index])
        y = np.zeros(x.shape[0])
        y[0] = 1

        clf = svm.LinearSVC(
            class_weight="balanced",
            verbose=False,
            max_iter=50000,
            tol=1e-4,
            C=0.1,
            dual=True,
        )
        clf.fit(x, y)

        similarities = clf.decision_function(x)
        sorted_ix = np.argsort(-similarities)

        # svm.LinearSVC in scikit-learn is non-deterministic.
        # if a text is the same as a query, there is no guarantee
        # the query will be in the first index.
        # this performs a simple swap, this works because anything
        # left of the 0 should be equivalent.
        zero_index = np.where(sorted_ix == 0)[0][0]
        if zero_index != 0:
            sorted_ix[0], sorted_ix[zero_index] = sorted_ix[zero_index], sorted_ix[0]

        denominator = np.max(similarities) - np.min(similarities) + 1e-6
        normalized_similarities = (similarities - np.min(similarities)) / denominator

        top_k_results = []
        top_k_scores = []
        for row in sorted_ix[1 : self.k + 1]:
            # print(normalized_similarities[row])  # stats
            if (
                self.relevancy_threshold is None
                or normalized_similarities[row] >= self.relevancy_threshold
            ):
                top_k_results.append(
                    Document(
                        page_content=self.texts[row - 1],
                        metadata=self.metadata[row - 1],
                    )
                )
                top_k_scores.append(normalized_similarities[row])
        return top_k_results, top_k_scores

    def get_relevant_documents_and_scores(
        self,
        query: str,
    ) -> Tuple[List[Document], List[float]]:
        return self._gs(query)

    async def _aget_relevant_documents(self, query: str) -> List[Document]:
        return self._gs(query)[0]

    def _get_relevant_documents(self, query: str) -> List[Document]:
        return self._gs(query)[0]



from openai import OpenAI, AzureOpenAI
from langchain.llms.base import LLM
import base64
from pathlib import Path
import openai as ai
from typing import Optional, List, Union, Any, Generator
from pydantic import BaseModel
import asyncio
from openai import AsyncOpenAI, AsyncAzureOpenAI


async def _async_get_completion(
    client: Union[AsyncAzureOpenAI, AsyncOpenAI],
    model: str,
    prompt: dict,
    max_tokens: int,
    temperature: float,
):
    response = await client.chat.completions.create(
        model=model,
        messages=prompt,
        max_tokens=max_tokens,
        temperature=temperature,
    )
    return response.choices[0].message.content


class AzureOpenAIClient(LLM):
    model_name: str = "gpt-4o"
    max_output_tokens: int = 1024
    temperature: float = 0.0
    api_key: str = ""
    api_type: str = "openai"
    api_base: str = ""
    api_version: str = "2023-05-15"
    client: Union[OpenAI, AzureOpenAI, None] = None

    def __init__(
        self,
        api_key: str,
        model_name: str = "gpt-4o",
        max_output_tokens: int = 1024,
        temperature: float = 0.0,
        api_type: str = "openai",
        api_base: str = "",
        api_version: str = "2023-05-15",
        **kwargs,
    ):
        super().__init__()
        self.model_name = model_name
        self.max_output_tokens = max_output_tokens
        self.temperature = temperature
        self.api_type = api_type
        self.api_key = api_key
        if api_type == "openai":
            self.client = OpenAI(
                api_key=api_key,
            )

        else:
            self.client = AzureOpenAI(
                azure_endpoint=api_base,
                azure_deployment=model_name,
                api_key=api_key,
                api_version=api_version,
            )
            self.api_base = api_base
            self.api_version = api_version

    @property
    def _llm_type(self) -> str:
        """return llm type

        Returns:
            str: llm type
        """
        return f"openai:{self.model_name}"

    def invoke(
        self,
        messages: Union[list, str],
        stop: Optional[List[str]] = None,
        verbose: bool = True,
        response_format: Union[dict, BaseModel, None] = None,
    ) -> str:
        """run llm and get the stream generator

        Args:
            prompt (str): _description_

        Yields:
            Generator: _description_
        """
        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]

        params = {
            "model": self.model_name,
            "messages": messages,
            "temperature": self.temperature,
            "max_tokens": self.max_output_tokens,
        }

        fully_text = ""
        if response_format is not None:
            params["response_format"] = response_format
            response = self.client.beta.chat.completions.parse(**params)
            message = response.choices[0].message

            return message.content
        else:
            params["stream"] = True
            response = self.client.chat.completions.create(**params)

        for chunk in response:
            if chunk.choices[0].delta.content:
                fully_text += chunk.choices[0].delta.content
                if verbose:
                    print(chunk.choices[0].delta.content, end="", flush=True)

        return fully_text

    def _call(
        self,
        messages: list,
        stop: Optional[List[str]] = None,
        verbose: bool = True,
        resposne_format: Union[dict, BaseModel, None] = None,
    ) -> str:
        return self.invoke(messages, stop, verbose, resposne_format)

    def stream(
        self,
        messages: Union[list, str],
        stop: Optional[List[str]] = None,
        verbose: bool = True,
        resposne_format: Union[dict, BaseModel, None] = None,
    ) -> Generator[Any, Any, Any]:
        """run llm and get the stream generator

        Args:
            prompt (str): _description_

        Yields:
            Generator: _description_
        """
        if isinstance(messages, str):
            messages = [{"role": "user", "content": messages}]

        params = {
            "model": self.model_name,
            "messages": messages,
            "temperature": self.temperature,
            "max_tokens": self.max_output_tokens,
            "stream": True,
        }
        fully_text = ""

        response = self.client.chat.completions.create(**params)

        for chunk in response:
            if chunk.choices[0].delta.content:
                yield chunk.choices[0].delta.content
                fully_text += chunk.choices[0].delta.content

        return

    def batch(
        self,
        inputs: list,
        stop: Optional[List[str]] = None,
        verbose: bool = True,
        resposne_format: Union[dict, BaseModel, None] = None,
    ) -> List[str]:
        messages_list = []
        for prompt in inputs:
            if isinstance(prompt, str):
                prompt = [{"role": "user", "content": prompt}]
            messages_list.append(prompt)

        async def process_batch(prompts):
            # 根據 api_type 決定用哪個 async client
            if self.api_type == "openai":
                client = AsyncOpenAI(api_key=self.api_key)
            else:
                client = AsyncAzureOpenAI(
                    azure_endpoint=self.api_base,
                    azure_deployment=self.model_name,
                    api_key=self.api_key,
                    api_version=self.api_version,
                )
            tasks = [
                _async_get_completion(
                    client,
                    self.model_name,
                    prompt,
                    self.max_output_tokens,
                    self.temperature,
                )
                for prompt in prompts
            ]
            return await asyncio.gather(*tasks)

        return asyncio.run(process_batch(messages_list))

    def generate(
        self,
        prompt: str,
        save_path: str = "./image.png",
        size: str = "auto",
        quality: str = "auto",
        moderation: str = "auto",
        background: str = "auto",
    ) -> Path:
        """generate image based on the user prompt.\n
        Args:
            prompt (str): _description_
            size (str, optional): _description_. Defaults to "auto".
            quality (str, optional): _description_. Defaults to "auto".
            output_format (str, optional): _description_. Defaults to "png".
            moderation (str, optional): _description_. Defaults to "auto".
            background (str, optional): _description_. Defaults to "auto".
        """
        if "dall" in self.model_name.lower():
            if size == "auto":
                size = "1024x1024"
            if quality == "auto":
                quality = "standard"
        # Convert save_path to Path object
        path = Path(save_path)
        allowed_exts = {"png", "jpeg", "webp"}

        if path.is_dir():
            # If it's a directory, append default filename
            path = path / "image.png"
            output_format = "png"
        else:
            ext = path.suffix.lower().lstrip(".")
            if ext not in allowed_exts:
                # Invalid or missing extension → default to .png
                path = path.with_suffix(".png")
                output_format = "png"
            else:
                output_format = ext
        params = {
            "model": self.model_name,
            "prompt": prompt,
            "size": size,
            "quality": quality,
            "output_format": output_format,
        }
        if "dall" in self.model_name.lower():
            # For DALL-E models, we need to set the response format to b64_json
            params["response_format"] = "b64_json"
        # Generate image
        result = self.client.images.generate(
            **params,
        )
        ## decode the image

        image_base64 = result.data[0].b64_json
        image_bytes = base64.b64decode(image_base64)

        ## save the image
        with open(path, "wb") as f:
            f.write(image_bytes)

        print(f"Image saved to {path}")

        return path

    def edit(
        self,
        prompt: str,
        images: str,
        save_path: str = "./image.png",
        size: str = "auto",
        quality: str = "auto",
        moderation: str = "auto",
        background: str = "auto",
    ) -> Path:
        """generate image based on the user prompt.\n
        Args:
            prompt (str): _description_
            size (str, optional): _description_. Defaults to "auto".
            quality (str, optional): _description_. Defaults to "auto".
            output_format (str, optional): _description_. Defaults to "png".
            moderation (str, optional): _description_. Defaults to "auto".
            background (str, optional): _description_. Defaults to "auto".
        """
        ai._azure._deployments_endpoints.add("/images/edits")  # hack fix openai bug
        ## process the image list
        images_source = []
        for image in images:
            if isinstance(image, str):
                # use Path to check if the image is exist
                image_path = Path(image)
                if not image_path.exists():
                    raise ValueError(f"Image {image} does not exist.")
                images_source.append(open(image, "rb"))
            elif isinstance(image, Path):
                if not image.exists():
                    raise ValueError(f"Image {image} does not exist.")
                images_source.append(open(image, "rb"))
            else:
                raise ValueError(f"Image {image} is not a valid path or file object.")

        # Convert save_path to Path object
        path = Path(save_path)
        allowed_exts = {"png", "jpeg", "webp"}

        if path.is_dir():
            # If it's a directory, append default filename
            path = path / "image.png"

        else:
            ext = path.suffix.lower().lstrip(".")
            if ext not in allowed_exts:
                # Invalid or missing extension → default to .png
                path = path.with_suffix(".png")

        # Generate image
        result = self.client.images.edit(
            image=images_source,
            model=self.model_name,
            prompt=prompt,
            size=size,
        )

        ## decode the image

        image_base64 = result.data[0].b64_json
        image_bytes = base64.b64decode(image_base64)

        ## save the image
        with open(path, "wb") as f:
            f.write(image_bytes)

        print(f"Image saved to {path}")

        return path



from langchain_chroma import Chroma
from chromadb.config import Settings
import logging
from pathlib import Path
from typing import Union, Callable
from langchain_core.embeddings import Embeddings
from akasha.utils.db.db_structure import (
    get_storage_directory,
    FILE_LAST_CHANGE_FILE_NAME,
)
from akasha.helper.base import get_embedding_type_and_name
import shutil
import gc
import json
from typing import List

_LANGCHAIN_DEFAULT_COLLECTION_NAME = "langchain"


def delete_documents_by_directory(
    directory_path: Union[str, Path],
    embeddings: Union[str, Embeddings, Callable],
    chunk_size: int,
) -> int:
    """delete the documents in the chroma db by directory path

    Args:
        directory_path (Union[str, Path]): _description_
        embeddings (Union[str, Embeddings, Callable]): _description_
        chunk_size (int): _description_

    Returns:
        int: _description_
    """

    try:
        embed_type, embed_name = get_embedding_type_and_name(embeddings)
        storage_directory = get_storage_directory(
            directory_path, chunk_size, embed_type, embed_name
        )

        shutil.rmtree(Path(storage_directory))
    except Exception as e:
        logging.warning(f"Error deleting directory {directory_path}: {e}")
        print(f"Error deleting directory {directory_path}: {e}")
        return 0

    return 1


def delete_documents_by_file(
    file_path: Union[Path, str],
    embeddings: Union[str, Embeddings, Callable],
    chunk_size: int,
) -> int:
    """delete the documents in the chroma db by file name

    Args:
        file_name (str): _description_
        embeddings (Union[str, Embeddings, Callable]): _description_
        chunk_size (int): _description_

    Returns:
        int:  the number of deleted documents
    """
    if isinstance(file_path, str):
        file_path = Path(file_path)
    doc_path = file_path.parent
    file_name = file_path.name
    embed_type, embed_name = get_embedding_type_and_name(embeddings)
    storage_directory = get_storage_directory(
        doc_path, chunk_size, embed_type, embed_name
    )

    # Retrieve all documents in the collection
    client_settings = Settings(
        is_persistent=True,
        persist_directory=storage_directory,
        anonymized_telemetry=False,
    )
    docsearch = Chroma(
        persist_directory=storage_directory, client_settings=client_settings
    )
    all_docs = docsearch._collection.get()
    tot_ids_len = len(all_docs["ids"])
    # Filter documents by metadata
    ids_to_delete = [
        doci
        for doci, docm in zip(all_docs["ids"], all_docs["metadatas"])
        if file_name in docm.get("source")
    ]
    # Delete documents by IDs
    if len(ids_to_delete) > 0:
        docsearch._collection.delete(ids=ids_to_delete)

        logging.info(
            f"Deleted {len(ids_to_delete)} documents with file_name: {file_name}"
        )
    else:
        logging.warning(f"No documents found with file_name: {file_name} to delete")
        print(f"No documents found with file_name: {file_name} to delete")

    if tot_ids_len == len(ids_to_delete) or tot_ids_len == 0:
        del all_docs
        gc.collect()
        docsearch._client.delete_collection(_LANGCHAIN_DEFAULT_COLLECTION_NAME)
        del docsearch
        gc.collect()
        shutil.rmtree(Path(storage_directory))
        logging.info(f"Deleted all documents in the directory: {doc_path}")
        print(f"Deleted all documents in the directory: {doc_path}")
        return tot_ids_len
    else:
        _delete_docs_built_time(storage_directory, [file_name])

    del docsearch, all_docs
    gc.collect()
    return len(ids_to_delete)


def delete_documents_from_chroma_by_file_name(chroma: Chroma, file_name: str) -> int:
    """delete the documents in the chroma db by file name

    Args:
        chroma (Chroma): _description_
        file_name (str): _description_

    Returns:
        int: the number of deleted documents
    """
    all_docs = chroma._collection.get()
    # Filter documents by metadata
    ids_to_delete = [
        doci
        for doci, docm in zip(all_docs["ids"], all_docs["metadatas"])
        if file_name in docm.get("source")
    ]
    # Delete documents by IDs
    if len(ids_to_delete) > 0:
        chroma._collection.delete(ids=ids_to_delete)
        logging.info(
            f"Deleted {len(ids_to_delete)} documents with file_name: {file_name}"
        )
    else:
        logging.info(f"No documents found with file_name: {file_name}")

    return len(ids_to_delete)


def _delete_docs_built_time(storage_directory: str, file_name: List[str]):
    storage_path = Path(storage_directory)

    # Path to the JSON file
    json_file_path = storage_path / FILE_LAST_CHANGE_FILE_NAME

    # Create the dictionary
    file_last_changed = {}

    # Check if the JSON file exists
    if json_file_path.exists():
        # Load the JSON file
        try:
            with open(json_file_path, "r", encoding="utf-8") as json_file:
                file_last_changed = json.load(json_file)
        except Exception as e:
            logging.warning(f"Error reading JSON file: {e}")
            print(f"Error reading JSON file: {e}")
    else:
        logging.warning(f"JSON file not found: {json_file_path}")
        print(f"JSON file not found: {json_file_path}")
        return

    # Update the dictionary
    for f_name in file_name:
        file_last_changed.pop(f_name, None)

    # Write the dictionary to the JSON file
    try:
        with open(json_file_path, "w", encoding="utf-8") as json_file:
            json.dump(file_last_changed, json_file, indent=4, ensure_ascii=False)
    except Exception as e:
        logging.warning(f"Error writing JSON file: {e}")
        print(f"Error writing JSON file: {e}")

    return



from setuptools import setup

# read the contents of your README file
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

install_requires = [
    "pypdf",
    "langchain-core>=0.3,<0.4",
    "langchain>=0.3,<0.4",
    "langchain-community>=0.3,<0.4",
    "langchain_openai>=0.1.0",
    "langchain-huggingface>=0.1.2",
    "langchain-chroma==0.2.2",
    "chromadb==0.4.14",
    "onnxruntime <= 1.20.0",
    "openai>=1.78.1",
    "tiktoken",
    "scikit-learn>=1.3.0",
    "jieba>=0.42.1",
    "opencc-python-reimplemented==0.1.7",
    "sentence-transformers>=3.1.1",
    "transformers>=4.45.0,<4.48",  # ==4.41.1
    "accelerate>=1.7.0",
    "tqdm>=4.65.0",
    "docx2txt==0.8",
    "rouge==1.0.1",
    "rouge-chinese==1.0.3",
    "bert-score==0.3.13",
    "click",
    "tokenizers>=0.19.1",
    "streamlit>=1.33.0",
    "streamlit_option_menu>=0.3.6",
    "rank_bm25",
    "unstructured",
    "python-pptx",
    "wikipedia",
    "numpy<2",
    "sentencepiece",
    "google-genai>=1.15.0",
    "anthropic",
    "tavily-python",
    "faiss-cpu",
    "langchain-mcp-adapters>=0.1.1",
    "mlflow",
    "setuptools",
]

### install different torch version###
install_requires.append("torch==2.2.0; platform_system=='Windows'")
install_requires.append("torch==2.0.1; platform_system=='Darwin'")
install_requires.append("torch==2.2.0; platform_system=='Linux'")

install_requires.append("torchvision==0.17.0; platform_system=='Windows'")
install_requires.append("torchvision==0.15.2; platform_system=='Darwin'")
install_requires.append("torchvision==0.17.0; platform_system=='Linux'")
setup(
    name="akasha_terminal",
    version="0.9.05",
    description="document QA(RAG) package using langchain and chromadb",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="chih chuan chang",
    url="https://github.com/iii-org/akasha",
    author_email="ccchang@iii.org.tw",
    install_requires=install_requires,
    extras_require={
        "llama-cpp": [
            "llama-cpp-python>=0.3.8",
        ],
        "peft": ["auto-gptq==0.3.1"],
    },
    packages=[
        "akasha",
        "akasha.agent",
        "akasha.interface",
        "cli",
        "akasha.eval",
        "akasha.helper",
        "akasha.RAG",
        "akasha.tools",
        "akasha.utils",
        "akasha.utils.models",
        "akasha.utils.db",
        "akasha.utils.prompts",
        "akasha.utils.search",
        "akasha.utils.search.retrievers",
    ],
    entry_points={"console_scripts": ["akasha = cli.glue:akasha"]},
    python_requires=">=3.10,<3.13",
)



from langchain_community.document_loaders import (
    PyPDFLoader,
    TextLoader,
    Docx2txtLoader,
    UnstructuredPowerPointLoader,
)
from langchain_community.document_loaders.csv_loader import CSVLoader
from langchain.schema import Document
from typing import Union, List
import warnings
import logging
import traceback
from akasha.helper.encoding import detect_encoding
from pathlib import Path
from .db_structure import TEXT_EXTENSIONS
from akasha.helper.crawler import get_text_from_url

warnings.filterwarnings("ignore", category=UserWarning, module="pypdf")
logging.basicConfig(level=logging.ERROR)


def load_file(file_path: str, extension: str) -> List[Document]:
    """get the content and metadata of a text file (.pdf, .docx, .md, .txt, .csv, .pptx) and return a Document object

    Args:
        file_path (str): the path of the text file\n
        extension (str): the extension of the text file\n

    Raises:
        Exception: if the length of doc is 0, raise error

    Returns:
        list: list of Docuemnts
    """
    try:
        if extension == "pdf" or extension == "PDF":
            docs = PyPDFLoader(file_path).load()
        elif extension == "docx" or extension == "DOCX":
            docs = Docx2txtLoader(file_path).load()
            for i in range(len(docs)):
                docs[i].metadata["page"] = i

        elif extension == "csv":
            encoding = detect_encoding(file_path)
            docs = CSVLoader(file_path, encoding=encoding).load()
            for i in range(len(docs)):
                docs[i].metadata["page"] = docs[i].metadata["row"]
                del docs[i].metadata["row"]
        elif extension == "pptx":
            docs = UnstructuredPowerPointLoader(file_path).load()
            for i in range(len(docs)):
                docs[i].metadata["page"] = i
        else:
            docs = TextLoader(file_path, encoding="utf-8").load()
            for i in range(len(docs)):
                docs[i].metadata["page"] = i
        if len(docs) == 0:
            raise Exception

        return docs
    except Exception as err:
        try:
            trace_text = traceback.format_exc()

            logging.warning(
                "\nLoad "
                + file_path
                + " failed, ignored.\n"
                + trace_text
                + "\n\n"
                + str(err)
            )
        except Exception:
            logging.warning("\nLoad file" + " failed, ignored.\n" + trace_text + "\n\n")
        return []


def load_directory(directory_path: Union[str, Path]) -> List[Document]:
    """get the content and metadata of text files in a directory and return a list of Document objects

    Args:
        directory_path (Union[str, Path]): the path of the directory\n

    Returns:
        list: list of Documents
    """

    ### get the chromadb directory name###
    if isinstance(directory_path, str):
        directory_path = Path(directory_path)

    files, docs = [], []

    for extension in TEXT_EXTENSIONS:
        files.extend(get_load_file_list(directory_path, extension))

    for file in files:
        whole_file_path = directory_path / file
        file_doc = load_file(whole_file_path, file.split(".")[-1])
        if file_doc == "" or len(file_doc) == 0:
            logging.warning(f"file {file} load failed or empty.\n\n")
            print(f"file {file} load failed or empty.\n\n")

        docs.extend(file_doc)
    return docs


def load_url(url: str) -> Document:
    """get the content and metadata of a url and return a Document object

    Args:
        url (str): the url\n

    Returns:
        Document: Document object
    """
    url_title, url_text = get_text_from_url(url)

    return Document(page_content=url_text, metadata={"title": url_title, "url": url})


def get_load_file_list(doc_path: Union[str, Path], extension: str = "pdf") -> list:
    """get the list of text files with extension type in doc_path directory

    Args:
        **doc_path (Union[str,Path])**: text files directory\n
        **extension (str, optional):** the extension type. Defaults to "pdf".\n

    Returns:
        list: list of filenames
    """
    dir = Path(doc_path)
    pdf_files = dir.glob("*." + extension)
    loaders = [file.name for file in pdf_files]

    return loaders



from akasha.utils.prompts.gen_prompt import (
    format_wrong_answer,
    format_sys_prompt,
    format_category_prompt,
    compare_question_prompt,
    default_doc_ask_prompt,
    format_question_query,
    format_llama_json,
)
from akasha.helper import call_model
from akasha.helper.base import extract_json, get_docs_length, get_doc_length
from akasha.helper.scores import get_bert_score, get_llm_score, get_rouge_score
from akasha.utils.prompts.format import (
    handle_score_table,
    handle_metrics,
    handle_params,
    handle_table,
)
from akasha.utils.atman import atman
from akasha.utils.search.search_doc import search_docs
from akasha.utils.base import (
    DEFAULT_CHUNK_SIZE,
    DEFAULT_EMBED,
    DEFAULT_MAX_INPUT_TOKENS,
    DEFAULT_MAX_OUTPUT_TOKENS,
    DEFAULT_MODEL,
    DEFAULT_SEARCH_TYPE,
)
from .base import get_non_repeat_rand_int, find_same_category

import numpy as np
from tqdm import tqdm
import json
import importlib
from collections import defaultdict

from typing import Union, Callable
import os
import datetime
import time
import traceback
import logging
from langchain_core.language_models.base import BaseLanguageModel
from langchain_core.embeddings import Embeddings
from langchain.schema import Document


def get_torch():
    ttorch = importlib.import_module("torch")
    return ttorch


def _generate_single_choice_question(
    doc_text: str,
    question: str,
    cor_ans: str,
    model,
    system_prompt: str,
    choice_num: int,
) -> list:
    """Based on gernerated question and answer, generate wrong answers for single choice question

    Args:
        **doc_text (str)**: the document text that used to generate question and answer\n
        **question (str)**: question generated from previous step\n
        **cor_ans (str)**: correct answer genereated from previous step\n
        **model (var)**: llm model\n
        **system_prompt (str)**: the system prompt that you assign special instruction to llm model, currently not be used in the function\n
        **choice_num (int)**: the number of options for each single choice question\n

    Raises:
        Exception: if the format of the response is not correct, raise exception

    Returns:
        list: choice_num of wrong answers and a correct answer, and the index of correct answer
    """
    res = []
    count = 0
    random_index = np.random.randint(choice_num)
    q_prompt = format_wrong_answer(choice_num - 1, doc_text, question, cor_ans)

    input_text = format_sys_prompt(system_prompt, q_prompt)
    response = call_model(model, input_text)

    ### separate the response into wrong answers ###
    try:
        process = response.split("錯誤答案：")
        process = process[1:]
        if len(process) != choice_num - 1:
            raise Exception("Answer Format Error")
    except Exception:
        try:
            process = response.split("：")[1:]
            if len(process) != choice_num - 1:
                raise Exception("Answer Format Error")
        except Exception:
            process = response.split(":")[1:]
    ### combine the wrong answers and correct answer into a single choice question ###
    for wrong_ans in process:
        if wrong_ans == "":
            continue
        elif count == random_index:
            res.append(str(count + 1) + "." + cor_ans.replace("\n", ""))
            count += 1

        wrong_ans = (
            str(count + 1) + "." + wrong_ans.replace("\n", "").replace("錯誤答案", "")
        )
        res.append(wrong_ans)
        count += 1

    if count < choice_num:
        res.append(str(count + 1) + "." + cor_ans.replace("\n", ""))

    res.append(str(random_index + 1))

    return res


class Model_Eval(atman):
    """class for implement evaluation of llm model, include auto_create_questionset and auto_evaluation."""

    def __init__(
        self,
        model: Union[str, BaseLanguageModel] = DEFAULT_MODEL,
        embeddings: Union[str, Embeddings] = DEFAULT_EMBED,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        search_type: Union[str, Callable] = DEFAULT_SEARCH_TYPE,
        max_input_tokens: int = DEFAULT_MAX_INPUT_TOKENS,
        max_output_tokens: int = DEFAULT_MAX_OUTPUT_TOKENS,
        question_type: str = "fact",
        question_style: str = "essay",
        temperature: float = 0.0,
        threshold: float = 0.0,
        language: str = "ch",
        record_exp: str = "",
        system_prompt: str = "",
        prompt_format_type: str = "auto",
        keep_logs: bool = False,
        use_chroma: bool = False,
        verbose: bool = False,
        env_file: str = "",
    ):
        super().__init__(
            model,
            embeddings,
            chunk_size,
            search_type,
            max_input_tokens,
            max_output_tokens,
            temperature,
            threshold,
            language,
            record_exp,
            system_prompt,
            keep_logs,
            verbose,
            use_chroma,
            env_file,
        )

        ### set argruments ###
        self.data_source = ""
        self.prompt_format_type = prompt_format_type
        self.question_type = question_type
        self.question_style = question_style
        self.question_num = 10
        self.eval_model = self.model_obj
        ### set variables ###

        self.docs = []
        self.prompt_tokens, self.prompt_length = [], []
        self.doc_tokens, self.doc_length = [], []
        self.question = []
        self.answer = []
        self.response = []
        self.score = {}

    def _add_basic_log(
        self,
        timestamp: str,
        fn_type: str,
        doc_range: int = -1,
        choice_num: int = -1,
        questionset_file: str = "",
    ) -> bool:
        if super()._add_basic_log(timestamp, fn_type) is False:
            return False

        self.logs[timestamp]["question_num"] = self.question_num
        self.logs[timestamp]["question_type"] = self.question_type
        self.logs[timestamp]["question_style"] = self.question_style

        if doc_range != -1 and choice_num != -1:
            self.logs[timestamp]["doc_range"] = doc_range
            self.logs[timestamp]["choice_num"] = choice_num
        else:
            self.logs[timestamp]["search_type"] = self.search_type
            self.logs[timestamp]["questionset_path"] = questionset_file
        return True

    def _add_result_log(self, timestamp: str, time: float) -> bool:
        """add post-process log to self.logs

        Args:
            timestamp (str): timestamp of this run
            time (float): spent time of this run
        """

        if super()._add_result_log(timestamp, time) is False:
            return False

        self.logs[timestamp]["prompt_length"] = self.prompt_length
        self.logs[timestamp]["prompt_tokens"] = self.prompt_tokens
        self.logs[timestamp]["doc_length"] = self.doc_length
        self.logs[timestamp]["doc_tokens"] = self.doc_tokens
        self.logs[timestamp]["question"] = self.question
        self.logs[timestamp]["answer"] = self.answer

        if "bert" in self.score:
            self.logs[timestamp]["bert"] = self.score["bert"]
            self.logs[timestamp]["rouge"] = self.score["rouge"]
            self.logs[timestamp]["llm_score"] = self.score["llm_score"]
        elif "correct_count" in self.score:
            self.logs[timestamp]["correct_count"] = self.score["correct_count"]

        return True

    def _display_info(self) -> bool:
        if self.verbose is False:
            return False
        print(f"Model: {self.model}, Embeddings: {self.embeddings}")
        print(f"Chunk size: {self.chunk_size}, Search type: {self.search_type}")

        return True

    def _display_info_fnl(self) -> bool:
        if self.verbose is False:
            return False
        print(
            f"Prompt tokens: {self.prompt_tokens}, Prompt length: {self.prompt_length}"
        )
        print(f"Doc tokens: {self.doc_tokens}, Doc length: {self.doc_length}\n\n")

        return True

    def _save_questionset(self, timestamp: str, output_file_path: str):
        """save questions and ref answers into txt file, and save the path of question set into logs

        Args:
            timestamp (str): the timestamp of the question set created function
            output_file_path (str): file name of the question set txt file, if not assign, use doc_path+datetime as the file name.
        """
        ### write question and answer into txt file, but first check if "questionset" directory exist or not, it not, first create it.
        ### for filename, if not assign, use doc_path+datetime as the file name.
        if not os.path.exists("questionset"):
            os.makedirs("questionset")

        if output_file_path == "":
            now = datetime.datetime.now()
            date_time_string = now.strftime("%Y-%m-%d_%H-%M-%S-%f")
            output_file_path = "questionset/" + str(date_time_string) + ".json"
        elif not output_file_path.endswith(".json"):
            output_file_path = output_file_path + ".json"

        data = {
            "question_style": self.question_style,
            "question_type": self.question_type,
            "question": self.question,
            "answer": self.answer,
        }

        with open(output_file_path, "w", encoding="utf-8") as f:
            json.dump(data, f, indent=4, ensure_ascii=False)

        print("\nquestion set saved in ", output_file_path, "\n\n")
        if self.keep_logs is True:
            self.logs[timestamp]["question"] = self.question
            self.logs[timestamp]["answer"] = self.answer
            self.logs[timestamp]["questionset_path"] = output_file_path

        return

    def _process_fact(
        self, response: str, doc_text: str, choice_num: int, source_file_name: str = ""
    ) -> bool:
        """parse the question and answer from the llm response, and save it into question and answer list.
        if can not parse the response, return False

        Args:
            response (str): llm generated response
            doc_text (str): the document text that used to generate question and answer
            choice_num (int): the number of options for each single choice question

        Returns:
            bool: if can not parse the response, return False
        """
        try:
            process = "".join(response.split("問題：")).split("答案：")
            if len(process) < 2:
                raise SyntaxError("Question Format Error")
        except Exception:
            process = "".join(response.split("問題:")).split("答案:")
            if len(process) < 2:
                return False

        process[0] = process[0].replace("根據文件", "")
        self.question.append(source_file_name + process[0])
        if self.question_style == "essay":
            self.answer.append(process[1])

        else:
            anss = _generate_single_choice_question(
                doc_text,
                process[0],
                process[1],
                self.model_obj,
                self.system_prompt,
                choice_num,
            )
            self.answer.append(anss)

        return True

    def _process_summary(self, response: str, doc_text: str) -> bool:
        """parse the question and answer from the llm response, and save it into question and answer list.
        if can not parse the response, return False

        Args:
            response (str): llm generated response
            doc_text (str): the document text that used to generate question and answer

        Returns:
            bool: if can not parse the response, return False
        """
        try:
            process = response.split("答案：")
            if len(process) < 2:
                raise SyntaxError("Question Format Error")
        except Exception:
            process = response.split("答案:")
            if len(process) < 2:
                return False

        self.question.append(doc_text.replace("\n", "") + "\n")
        self.answer.append(process[-1].replace("\n", ""))

        return True

    def _process_response(
        self, response: str, doc_text: str, choice_num: int, source_file_name: str = ""
    ):
        """process the response from llm model, and generate question and answer pair
        based on the question type and question style
        """
        if self.question_type.lower() in [
            "summary",
            "sum",
            "summarization",
            "summarize",
            "summaries",
            "摘要",
        ]:
            return self._process_summary(response, doc_text)

        return self._process_fact(response, doc_text, choice_num, source_file_name)

    def _create_compare_questionset(self, choice_num: int, output_file_path: str):
        """create compare question set, first randomly select documents and label the category of some proper nouns in the documents,
        can use the documents and proper nouns that have same category to generate compare question and answer pair.

        Args:
            choice_num (int): the number of options for each single choice question
            output_file_path (str): file name of the question set txt file, if not assign, use doc_path+datetime as the file name.

        Returns:
            _type_: _description_
        """
        ## set local variables ##
        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        start_time = time.time()
        doc_range = 1
        cate_threshold = 3
        vis_doc_range = set()
        self.doc_tokens, self.doc_length = [], []
        self.question, self.answer, self.docs = [], [], []
        table = {}

        ## add logs ##

        self._add_basic_log(timestamp, "create_questionset", doc_range, choice_num)

        texts = [pg_content for pg_content in self.db.docs]
        metadata = [metadata for metadata in self.db.metadatas]

        progress = tqdm(total=self.question_num, desc=f"Create Q({self.question_type})")
        print("\n")
        regenerate_limit = self.question_num
        category = defaultdict(list)
        set_category = defaultdict(set)
        for i in range(self.question_num):
            docs = []
            print(" ")
            progress.update(1)
            print("\n")

            ### start to random choose text, and use text to add different category things into dictionary,###
            ### if the value of any category large than category threshold, then we can use the category to create compare question ###
            compare_resource = find_same_category(category, cate_threshold)
            while not isinstance(compare_resource, list):
                random_index = get_non_repeat_rand_int(
                    vis_doc_range, len(texts) - doc_range, doc_range
                )
                doc_text = texts[random_index]
                docs.append(
                    Document(
                        page_content=texts[random_index],
                        metadata=metadata[random_index],
                    )
                )
                count = 3
                try:
                    ## ask model to get category & nouns, add to category ##
                    category_prompt = format_category_prompt(doc_text, self.language)
                    response = call_model(self.model_obj, category_prompt)

                    json_response = extract_json(response)
                    if json_response is None:
                        raise Exception("Response Format Error")

                    for k, v in json_response.items():
                        if k not in set_category[v]:
                            category[v].append([k, doc_text])
                            set_category[v].add(k)
                    compare_resource = find_same_category(category, cate_threshold)

                except Exception as e:
                    print("error during generate categories\n", e)
                    count -= 1
                    if count == 0:
                        break
            topic, nouns, used_texts = compare_resource
            used_texts = "\n".join(set(used_texts))
            nouns = ", ".join(nouns)
            try:
                q_prompt = compare_question_prompt(
                    self.question_style, topic, nouns, used_texts
                )
                response = call_model(self.model_obj, q_prompt)

                if not self._process_response(response, used_texts, choice_num, ""):
                    raise Exception(f"Question Format Error, got {response}")

                self.doc_length.append(get_docs_length(self.language, docs))
                self.doc_tokens.append(self.model_obj.get_num_tokens(used_texts))
                self.prompt_length.append(get_doc_length(self.language, q_prompt))
                self.prompt_tokens.append(self.model_obj.get_num_tokens(q_prompt))
                self.docs.extend(docs)

            except Exception as e:
                if regenerate_limit > 0:
                    regenerate_limit -= 1
                    i -= 1
                    progress.update(-1)
                    logging.warning(f"{e}.\n\n Regenerate\n")
                    continue
                else:
                    logging.error(f"{e}.\n\n Stop\n")
                    raise e

            # remove the category from category dictionary
            del category[topic]
            # remove the category from set_category dictionary
            del set_category[topic]

            # aiido upload
            table = self._update_table(table, docs)

        progress.close()  # end running llm progress bar

        end_time = time.time()

        ### record logs ###
        self._upload_logs(
            end_time - start_time, table, sum(self.doc_length), sum(self.doc_tokens)
        )
        self._add_result_log(timestamp, end_time - start_time)

        self._save_questionset(timestamp, output_file_path)

        return self.question, self.answer

    def _eval_get_res_fact(
        self,
        question: Union[str, list],
        answer: str,
        timestamp: str,
        retrivers_list: list,
    ) -> dict:
        """generate fact resposne from the question, can evaluate with reference answer

        Args:
            question (Union[str, list]): if it's single_choice, it should be a list(include options), else it should be a string of question
            answer (str): the reference answer of the question
            timestamp (str): the timestamp of the auto evaluation function

        Returns:
            dict: evaluation result
        """

        ### format question ###
        if self.question_style.lower() == "essay":
            query = question
            prod_sys = self.system_prompt + default_doc_ask_prompt(self.language)
            query_with_prompt = question

        else:
            prod_sys = self.system_prompt
            query, ans = format_question_query(question, answer)
            query_with_prompt = format_llama_json(query)

        ### get docs ###

        query_len = get_doc_length(self.language, query_with_prompt)
        query_tokens = self.model_obj.get_num_tokens(query_with_prompt)

        self.docs, doc_length, doc_tokens = search_docs(
            retrivers_list,
            query,
            self.model,
            self.max_input_tokens - query_tokens,
            self.search_type,
            self.language,
        )
        if self.doc_tokens == 0:
            print(
                "Warning: Unable to retrieve any documents, possibly due to insufficient remaining tokens.\n\n"
            )
            self.docs = []

        ### ask llm ###

        try:
            self._display_docs()
            intput_text = format_sys_prompt(
                prod_sys + self._format_docs(),
                query_with_prompt,
                self.prompt_format_type,
                self.model,
            )
            response = call_model(self.model_obj, intput_text)
            self.response.append(response)
            self.doc_length.append(doc_length)
            self.doc_tokens.append(doc_tokens)
            self.prompt_length.append(query_len)
            self.prompt_tokens.append(query_tokens)
        except Exception as e:
            traceback.print_exc()
            # response = ["running model error"]
            torch = get_torch()
            torch.cuda.empty_cache()
            logging.error(f"running model error\n {e}")
            raise e

        if self.question_style.lower() == "essay":
            if self.verbose:
                print("Question: ", question, "\n\n")
                print("Reference Answer: ", answer, "\n\n")
                print("Generated Response: ", response, "\n\n")

            self.score["bert"].append(get_bert_score(response, answer, self.language))
            self.score["rouge"].append(get_rouge_score(response, answer, self.language))
            self.score["llm_score"].append(
                get_llm_score(response, answer, self.eval_model, "auto")
            )

            new_table = handle_table(
                question + "\nAnswer:  " + answer, self.docs, response
            )
            new_table = handle_score_table(
                new_table,
                self.score["bert"][-1],
                self.score["rouge"][-1],
                self.score["llm_score"][-1],
            )

        else:
            if self.verbose:
                print("Question: ", question, "\n\n")
                print("Reference Answer: ", ans, "\n\n")
                print("Generated Response: ", response, "\n\n")

            new_table = handle_table(query + "\nAnswer:  " + ans, self.docs, response)
            new_table = {}
            result = extract_result(response)

            if str(ans).replace(" ", "") in str(result):
                self.score["correct_count"] += 1

        return new_table

    def _eval_get_res_summary(self, sum_doc: str, answer: str, timestamp: str) -> dict:
        """generate summary resposne from the question, can evaluate with reference answer

        Args:
            sum_doc (str): text that need to be summarized
            answer (str): the reference answer of the summary
            timestamp (str): the timestamp of the auto evaluation function

        Returns:
            dict: evaluation result
        """

        prompt = "請對以下文件進行摘要: "
        intput_text = format_sys_prompt(
            self.system_prompt,
            prompt + "\n\n" + sum_doc,
            self.prompt_format_type,
            self.model,
        )

        self.docs = [Document(page_content=sum_doc, metadata={"source": "", "page": 0})]

        try:
            response = call_model(self.model_obj, intput_text)
            self.response.append(response)
            self.doc_length.append(get_doc_length(self.language, sum_doc))
            self.doc_tokens.append(self.model_obj.get_num_tokens(sum_doc))
            self.prompt_length.append(8)
            self.prompt_tokens.append(15)

        except Exception as e:
            traceback.print_exc()
            # response = ["running model error"]
            torch = get_torch()
            torch.cuda.empty_cache()
            logging.error(f"running model error\n {e}")
            raise e

        if self.verbose:
            print("Question: ", prompt + "\n" + sum_doc, "\n\n")
            print("Reference Answer: ", answer, "\n\n")
            print("Generated Response: ", response, "\n\n")

        self.score["bert"].append(get_bert_score(response, answer, self.language))
        self.score["rouge"].append(get_rouge_score(response, answer, self.language))
        self.score["llm_score"].append(
            get_llm_score(response, answer, self.eval_model, "auto")
        )

        # new_table = akasha.format.handle_table(prompt + "\nAnswer:  " + answer,
        #                                        self.docs, response)
        # new_table = akasha.format.handle_score_table(
        #     new_table,
        #     self.score["bert"][-1],
        #     self.score["rouge"][-1],
        #     self.score["llm_score"][-1],
        # )
        new_table = {}

        return new_table

    def _eval_get_res(
        self,
        question: Union[list, str],
        answer: str,
        timestamp: str,
        retrivers_list: list,
    ) -> dict:
        """separate the question type and call different function to generate response

        Args:
            question (str): the question that need to be evaluated
            answer (str): the reference answer of the question
            timestamp (str): the timestamp of the auto evaluation function

        Returns:
            dict: _description_
        """
        if (
            self.question_type.lower()
            in ["fact", "facts", "factoid", "factoids", "事實"]
            or self.question_type.lower()
            in ["irre", "irrelevant", "irrelevance", "無關"]
            or self.question_type.lower()
            in ["compared", "compare", "comparison", "comparisons", "比較"]
        ):
            return self._eval_get_res_fact(question, answer, timestamp, retrivers_list)

        elif self.question_type.lower() in [
            "summary",
            "sum",
            "summarization",
            "summarize",
            "summaries",
            "摘要",
        ]:
            return self._eval_get_res_summary(question, answer, timestamp)

    def _update_table(self, table: dict, docs: list) -> dict:
        """add data into table dictionary for aiido

        Args:
            table (dict): _description_
            docs (list): _description_

        Returns:
            dict: _description_
        """
        if self.record_exp == "":
            return {}

        new_table = handle_table(self.question[-1], docs, self.answer[-1])
        for key in new_table:
            if key not in table:
                table[key] = []
            table[key].append(new_table[key])
        return table

    def _upload_logs(
        self,
        tot_time: float,
        table: dict,
        doc_length: int = 0,
        doc_tokens: int = 0,
        from_eval: bool = False,
    ) -> str:
        if self.record_exp == "":
            return "no record_exp assigned, so no logs uploaded"

        params = handle_params(
            self.model,
            self.language,
            self.search_type,
            self.threshold,
            self.embeddings,
            self.chunk_size,
        )
        metrics = handle_metrics(doc_length, tot_time, doc_tokens)
        if from_eval:
            if self.question_style.lower() == "essay":
                avg_bert = round(sum(self.score["bert"]) / len(self.score["bert"]), 3)
                avg_rouge = round(
                    sum(self.score["rouge"]) / len(self.score["rouge"]), 3
                )
                avg_llm_score = round(
                    sum(self.score["llm_score"]) / len(self.score["llm_score"]), 3
                )

                metrics["avg_bert"] = avg_bert
                metrics["avg_rouge"] = avg_rouge
                metrics["avg_llm_score"] = avg_llm_score
            else:
                correct_rate = self.score["correct_count"] / self.question_num
                metrics["correct_rate"] = correct_rate

        from akasha.utils.upload import aiido_upload

        aiido_upload(self.record_exp, params, metrics, table)

        return "logs uploaded"


def extract_result(response: str):
    """to prevent the output of llm format is not what we want, try to extract the answer (digit) from the llm output

    Args:
        **response (str)**: llm output\n

    Returns:
        int: digit of answer
    """
    try:
        res = extract_json(response)
        # res = str(json.loads(response)["ans"]).replace(" ", "")
        if res is None:
            raise Exception("can not find the json format in the response")
        res = res["ans"]
    except Exception:
        res = -1
        for c in response:
            if c.isdigit():
                res = c

                break
    return res



import click
import uvicorn
import akasha as ak


@click.group()
def akasha():
    pass


@click.command()
@click.option(
    "--data_source",
    "-d",
    help="document directory path, or url, parse all .txt, .pdf, .docx files in the directory",
    required=True,
)
@click.option("--prompt", "-p", help="prompt you want to ask to llm", required=True)
@click.option(
    "--embeddings",
    "-e",
    default="openai:text-embedding-ada-002",
    help="embeddings for storing the documents",
)
@click.option(
    "--chunk_size", "-c", default=1000, help="chunk size for storing the documents"
)
@click.option(
    "--model",
    "-m",
    default="openai:gpt-3.5-turbo",
    help="llm model for generating the response",
)
@click.option(
    "--language",
    "-l",
    default="ch",
    help="language for the documents, default is 'ch' for chinese",
)
@click.option(
    "--search_type",
    "-s",
    default="auto",
    help="search type for the documents, include auto, knn, svm, bm25",
)
@click.option(
    "--record_exp",
    "-r",
    default="",
    help="input the experiment name if you want to record the experiment using aiido",
)
@click.option(
    "--system_prompt", "-sys", default="", help="system prompt for the llm model"
)
@click.option(
    "--max_input_tokens", "-md", default=3000, help="max token for the llm model input"
)
def rag(
    data_source: str,
    prompt: str,
    embeddings: str,
    chunk_size: int,
    model: str,
    language: str,
    search_type: str,
    record_exp: str,
    system_prompt: str,
    max_input_tokens: int,
):
    gr = ak.RAG(
        verbose=False,
        embeddings=embeddings,
        chunk_size=chunk_size,
        model=model,
        language=language,
        search_type=search_type,
        record_exp=record_exp,
        system_prompt=system_prompt,
        max_input_tokens=max_input_tokens,
    )
    res = gr(data_source, prompt)

    print(res)

    del gr


@click.command()
@click.option(
    "--data_source",
    "-d",
    help="document directory path, or urls, parse all .txt, .pdf, .docx files in the directory",
    required=True,
)
@click.option(
    "--embeddings",
    "-e",
    default="openai:text-embedding-ada-002",
    help="embeddings for storing the documents",
)
@click.option(
    "--chunk_size", "-c", default=1000, help="chunk size for storing the documents"
)
@click.option(
    "--model",
    "-m",
    default="openai:gpt-3.5-turbo",
    help="llm model for generating the response",
)
@click.option(
    "--language",
    "-l",
    default="ch",
    help="language for the documents, default is 'ch' for chinese",
)
@click.option(
    "--search_type",
    "-s",
    default="merge",
    help="search type for the documents, include merge, svm, mmr, tfidf",
)
@click.option(
    "--system_prompt", "-sys", default="", help="system prompt for the llm model"
)
@click.option(
    "--max_input_tokens", "-md", default=3000, help="max token for the llm model input"
)
def keep_rag(
    data_source: str,
    embeddings: str,
    chunk_size: int,
    model: str,
    language: str,
    search_type: str,
    system_prompt: str,
    max_input_tokens: int,
):
    import akasha.helper as helper
    from langchain.chains.question_answering import load_qa_chain
    import akasha.utils.db as dd
    from akasha.utils.search.retrievers.base import get_retrivers
    from akasha.utils.search.search_doc import search_docs

    embeddings = helper.handle_embeddings(embeddings, False)
    model_name = model
    model = helper.handle_model(model, False)

    db, ign = dd.process_db(data_source, embeddings, chunk_size)

    if db is None:
        info = "document path not exist\n"
        print(info)
        return ""

    user_input = click.prompt('Please input your question(type "exit()" to quit) ')
    retrivers_list = get_retrivers(db, embeddings, 0.0, search_type)

    while user_input != "exit()":
        docs, docs_len, tokens = search_docs(
            retrivers_list,
            user_input,
            model_name,
            max_input_tokens,
            search_type,
            language,
        )
        if docs is None:
            docs = []

        chain = load_qa_chain(llm=model, chain_type="stuff", verbose=False)

        res = chain.run(input_documents=docs, question=system_prompt + user_input)
        res = helper.sim_to_trad(res)

        print("Response: ", res)
        print("\n\n")
        user_input = click.prompt('Please input your question(type "exit()" to quit) ')

    del db, model, embeddings


@click.command()
@click.option(
    "--data_source",
    "-d",
    help="document directory path, parse all .txt, .pdf, .docx files in the directory",
    required=True,
)
@click.option(
    "-question_num", "-qn", default=10, help="number of questions you want to generate"
)
@click.option(
    "-question_type",
    "--qt",
    default="essay",
    help="question type, include single_choice, essay",
)
@click.option(
    "--embeddings",
    "-e",
    default="openai:text-embedding-ada-002",
    help="embeddings for storing the documents",
)
@click.option(
    "--chunk_size", "-c", default=1000, help="chunk size for storing the documents"
)
@click.option(
    "--language",
    "-l",
    default="ch",
    help="language for the documents, default is 'ch' for chinese",
)
@click.option(
    "--search_type",
    "-s",
    default="merge",
    help="search type for the documents, include merge, svm, mmr, tfidf",
)
@click.option(
    "--record_exp",
    "-r",
    default="",
    help="input the experiment name if you want to record the experiment using aiido",
)
def create_questionset(
    data_source: str,
    question_num: int,
    question_type: str,
    embeddings: str,
    chunk_size: int,
    language: str,
    search_type: str,
    record_exp: str,
):
    model = "openai:gpt-3.5-turbo"
    eva = ak.eval(
        model=model,
        embeddings=embeddings,
    )
    eva.create_questionset(
        data_source,
        question_num,
        question_type=question_type,
        chunk_size=chunk_size,
        verbose=False,
        language=language,
        search_type=search_type,
        record_exp=record_exp,
    )

    del eva


@click.command()
@click.option(
    "--question_path",
    "-qp",
    help="document directory path, parse all .txt, .pdf, .docx files in the directory",
    required=True,
)
@click.option(
    "--data_source",
    "-d",
    help="document directory path, parse all .txt, .pdf, .docx files in the directory",
    required=True,
)
@click.option(
    "--question_type",
    "-qt",
    default="essay",
    help="question type, include single_choice, essay",
)
@click.option(
    "--embeddings",
    "-e",
    default="openai:text-embedding-ada-002",
    help="embeddings for storing the documents",
)
@click.option(
    "--chunk_size", "-c", default=1000, help="chunk size for storing the documents"
)
@click.option(
    "--model",
    "-m",
    default="openai:gpt-3.5-turbo",
    help="llm model for generating the response",
)
@click.option("--topk", "-k", default=2, help="select topK relevant documents")
@click.option(
    "--language",
    "-l",
    default="ch",
    help="language for the documents, default is 'ch' for chinese",
)
@click.option(
    "--search_type",
    "-s",
    default="merge",
    help="search type for the documents, include merge, svm, mmr, tfidf",
)
@click.option(
    "--record_exp",
    "-r",
    default="",
    help="input the experiment name if you want to record the experiment using aiido",
)
@click.option(
    "--max_input_tokens", "-md", default=3000, help="max token for the llm model input"
)
def evaluation(
    question_path: str,
    data_source: str,
    question_type: str,
    embeddings: str,
    chunk_size: int,
    model: str,
    language: str,
    search_type: str,
    record_exp: str,
    max_input_tokens: int,
):
    eva = ak.eval(model=model, embeddings=embeddings)

    if question_type.lower() == "single_choice":
        cor_rate, tokens = eva.evaluation(
            question_path,
            data_source,
            question_type=question_type,
            chunk_size=chunk_size,
            verbose=False,
            language=language,
            search_type=search_type,
            record_exp=record_exp,
            max_input_tokens=max_input_tokens,
        )

        print("correct rate: ", cor_rate)
        print("total tokens: ", tokens)

    else:
        avg_bert, avg_rouge, avg_llm, tokens = eva.evaluation(
            question_path,
            data_source,
            question_type=question_type,
            chunk_size=chunk_size,
            verbose=False,
            language=language,
            search_type=search_type,
            record_exp=record_exp,
            max_input_tokens=max_input_tokens,
        )

        print("avg bert score: ", avg_bert)
        print("average rouge score: ", avg_rouge)
        print("avg llm score: ", avg_llm)
        print("total tokens: ", tokens)


@click.command("toy", short_help="simple toy for akasha")
def ui():
    import os
    import sys
    import site
    from streamlit import config as _config
    from streamlit.web import cli as stcli

    # make a folder `docs/Default`
    if not os.path.exists("docs") or not os.path.exists(
        os.path.join("docs", "Default")
    ):
        os.makedirs(os.path.join(".", "docs", "Default"))
    else:
        pass

    if not os.path.exists("docs"):
        os.makedirs(os.path.join(".", "docs", "Default"))
    else:
        pass

    # make a folder `model`
    if not os.path.exists("model"):
        os.makedirs("model")
    else:
        pass

    # find the location of ui.py
    site_packages_dirs = site.getsitepackages()
    for dir in site_packages_dirs:
        if dir.endswith("site-packages"):
            target_dir = dir
            break
        else:
            target_dir = "."

    # run streamlit web service by ui.py
    _config.set_option("server.headless", True)

    ui_py_file = os.path.join(target_dir, "akasha", "ui.py")
    # streamlit.web.bootstrap.run(ui_py_file, "", [], [])
    sys.argv = ["streamlit", "run", ui_py_file]
    sys.exit(stcli.main())


@click.command("api", short_help="simple api for akasha")
@click.option(
    "--workers",
    "-w",
    default=2,
    help="Number of workers to use",
)
@click.option(
    "--host",
    "-h",
    default="0.0.0.0",
    help="Host to run the FastAPI server on",
)
@click.option(
    "--port",
    "-p",
    default=8000,
    help="Port to run the FastAPI server on",
)
def start_fastapi(workers: int, host: str, port: str):
    uvicorn.run("akasha.api:app", host=host, port=port, workers=workers)


akasha.add_command(keep_rag)
akasha.add_command(rag)
akasha.add_command(create_questionset)
akasha.add_command(evaluation)
akasha.add_command(ui)
akasha.add_command(start_fastapi)

if __name__ == "__main__":
    akasha()



from anthropic import Anthropic
from langchain.llms.base import LLM
from typing import Dict, List, Any, Optional, Generator, Union
from pydantic import Field
import concurrent.futures


class anthropic_model(LLM):
    max_token: int = 4096
    max_output_tokens: int = 1024
    temperature: float = 0.01
    top_p: float = 0.95
    history: list = []
    model: Anthropic = Field(default=None)
    model_name: str = "claude-3-5-sonnet-20241022"

    def __init__(
        self, model_name: str, api_key: str, temperature: float = 0.0, **kwargs
    ):
        """define custom model, input func and temperature

        Args:
            **func (Callable)**: the function return response from llm\n
        """
        super().__init__()
        self.model = Anthropic(api_key=api_key)
        self.model_name = model_name
        self.temperature = temperature
        if "max_output_tokens" in kwargs:
            self.max_output_tokens = kwargs["max_output_tokens"]

        #### gcp vertex (need credentials) ####

        # from anthropic import AnthropicVertex

        # client = AnthropicVertex(region=os.environ['GCP_LOCATION'],
        #                         project_id=os.environ['GCP_PROJECT_ID'])

        # with client.messages.stream(
        #         max_tokens=1024,
        #         messages=[{
        #             "role": "user",
        #             "content": "Send me a recipe for banana bread.",
        #         }],
        #         model="claude-3-sonnet@20240229",
        # ) as stream:
        #     for text in stream.text_stream:
        #         print(text)

    @property
    def _llm_type(self) -> str:
        """return llm type

        Returns:
            str: llm type
        """
        return f"anthropic:{self.model_name}"

    def stream(
        self, prompt: Union[str, List[Dict[str, Any]]], stop: Optional[List[str]] = None
    ) -> Generator:
        """run llm and get the stream generator

        Args:
            prompt (str): _description_

        Yields:
            Generator: _description_
        """
        if isinstance(prompt, str):
            prompt = [{"role": "user", "content": prompt}]

        with self.model.messages.stream(
            max_tokens=self.max_output_tokens,
            messages=prompt,
            model=self.model_name,
            stop_sequences=stop,
            temperature=self.temperature,
            top_p=self.top_p,
        ) as stream:
            for text in stream.text_stream:
                yield text

        return

    def _call(
        self,
        prompt: Union[str, List[Dict[str, Any]]],
        stop: Optional[List[str]] = None,
        verbose=True,
    ) -> str:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """

        ret = ""
        if isinstance(prompt, str):
            prompt = [{"role": "user", "content": prompt}]

        with self.model.messages.stream(
            max_tokens=self.max_output_tokens,
            messages=prompt,
            model=self.model_name,
            stop_sequences=stop,
            temperature=self.temperature,
            top_p=self.top_p,
        ) as stream:
            for text in stream.text_stream:
                ret += text
                if verbose:
                    print(text, end="", flush=True)

        return ret

    def _invoke_helper(self, args):
        messages, stop, verbose = args
        return self._call(messages, stop, verbose)

    def batch(self, prompt: List[str], stop: Optional[List[str]] = None) -> List[str]:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        # Number of threads should not exceed the number of prompts
        num_threads = min(
            len(prompt), concurrent.futures.thread.ThreadPoolExecutor()._max_workers
        )

        with concurrent.futures.ThreadPoolExecutor(max_workers=num_threads) as executor:
            results = list(
                executor.map(
                    self._invoke_helper, [(message, stop, False) for message in prompt]
                )
            )
        return results

    def invoke(
        self, messages: list, stop: Optional[List[str]] = None, verbose: bool = True
    ) -> str:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        return self._call(messages, stop, verbose)

    def invoke_stream(
        self, messages: list, stop: Optional[List[str]] = None
    ) -> Generator:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        return self.stream(messages, stop)

    def call_image(self, prompt: list, stop: Optional[List[str]] = None) -> str:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """

        messages = self.model.messages.create(
            max_tokens=self.max_output_tokens,
            messages=prompt,
            model=self.model_name,
            stop_sequences=stop,
            temperature=self.temperature,
            top_p=self.top_p,
        )

        return messages.content[0].text

    def count_tokens(self, prompt: Union[list, str]) -> int:
        """caculate the token count

        Args:
            prompt (Union[list,str]): _description_

        Returns:
            int: _description_
        """
        if isinstance(prompt, str):
            input_text = [{"role": "user", "content": prompt}]

        token_count = self.model.beta.messages.count_tokens(
            model=self.model_name, messages=input_text
        )

        return token_count.input_tokens



from typing import Union, List, Tuple, Callable, Optional
from pathlib import Path
from langchain_core.embeddings import Embeddings
from akasha.utils.db.db_structure import dbs, get_storage_directory, is_url
from akasha.utils.db.extract_db import extract_db_by_file
from akasha.utils.db.create_db import (
    create_directory_db,
    create_single_file_db,
    create_webpage_db,
)
from akasha.helper import separate_name
from akasha.helper.handle_objects import handle_embeddings_and_name
from langchain_chroma import Chroma
from chromadb.config import Settings
import logging
import gc
from collections import defaultdict
from tqdm import tqdm


def process_db(
    data_source: Union[List[Union[str, Path]], Union[Path, str]],
    embeddings: Union[str, Embeddings, Callable],
    chunk_size: int,
    verbose: bool = False,
    env_file: str = "",
) -> Tuple[dbs, List[str]]:
    """create and load dbs object from data_source

    Args:
        data_source (Union[List[Union[str,Path]], Union[Path,str]]): _description_

    Returns:
        Tuple[dbs, List[str]]: return dbs object and list of ignored files
    """
    ignored_files = []
    embeddings, embeddings_name = handle_embeddings_and_name(
        embeddings, False, env_file
    )
    tot_db = dbs()

    direct_list = []
    files_dict = defaultdict(list)
    suc_files = []
    link_list = []
    file_count = 0
    if not isinstance(data_source, list):
        data_source = [data_source]

    for data_path in data_source:
        if isinstance(data_path, str):
            if data_path == "":
                continue
            if is_url(data_path):
                link_list.append(data_path)
                continue
            data_path = Path(data_path)

        if not data_path.exists():
            logging.warning(f"File {data_path} not found")
            print(f"File {data_path} not found")
            ignored_files.append(data_path)
            continue

        ### divide data_path into dir, files and links ###
        if data_path.is_dir():
            direct_list.append(data_path)

        else:  # if files
            files_dict[data_path.parent.__str__()].append(data_path)
            file_count += 1

    ### load dbs object based on directories ###
    for data_path in direct_list:
        try:
            is_suc, cur_ignores = create_directory_db(
                data_path,
                embeddings,
                chunk_size,
                env_file=env_file,
            )
            ignored_files.extend(cur_ignores)

            if is_suc:
                new_dbs = load_directory_db(
                    data_path,
                    embeddings,
                    chunk_size,
                )
                tot_db.merge(new_dbs)
        except Exception as e:
            logging.warning(f"Error loading directory {data_path}: {e}")
            print(f"Error loading directory {data_path}: {e}")
            continue

    ### load dbs object based on files ###
    for parent_dir, file_list in files_dict.items():
        progress = tqdm(total=len(file_list), desc=f"db {parent_dir}")

        for data_path in file_list:
            progress.update(1)
            try:
                is_suc = create_single_file_db(
                    data_path, embeddings, chunk_size, env_file=env_file
                )

                if is_suc:
                    suc_files.append(data_path)
                else:
                    ignored_files.append(data_path)
            except Exception as e:
                logging.warning(f"Error loading file {data_path}: {e}")
                print(f"Error loading file {data_path}: {e}")
                ignored_files.append(data_path)
                continue
        progress.close()

    if len(suc_files) > 0:
        new_dbs = load_files_db(suc_files, embeddings, chunk_size)
        tot_db.merge(new_dbs)

    ### load dbs object based on links ###
    if len(link_list) > 0:
        progress = tqdm(total=len(link_list), desc="db http")
        for url in link_list:
            progress.update(1)
            try:
                is_suc = create_webpage_db(
                    url,
                    embeddings,
                    chunk_size,
                    env_file=env_file,
                )

                if is_suc:
                    new_dbs = load_directory_db(
                        url,
                        embeddings,
                        chunk_size,
                    )
                    tot_db.merge(new_dbs)
                else:
                    ignored_files.append(url)
            except Exception as e:
                logging.warning(f"Error loading directory {data_path}: {e}")
                print(f"Error loading directory {data_path}: {e}")
                continue
        progress.close()

    ### print the information ###
    _display_db_num(len(direct_list), file_count, len(link_list), verbose)

    return tot_db, ignored_files


def load_db_by_chroma_name(
    chroma_name_list: Union[List[Union[str, Path]], Union[Path, str]],
) -> Tuple[dbs, List[str]]:
    """load dbs object from chroma db name

    Args:
        chroma_name_list (List[Union[str, Path]]): _description_

    Returns:
        Tuple[dbs, List[str]]: return dbs object and list of ignored files
    """

    if not isinstance(chroma_name_list, list):
        chroma_name_list = [chroma_name_list]

    ignored_files = []
    tot_db = dbs()
    for chroma_name in chroma_name_list:
        if isinstance(chroma_name, Path):
            chroma_name = chroma_name.__str__()

        if not Path(chroma_name).exists():
            logging.warning(f"File {chroma_name} not found")
            print(f"File {chroma_name} not found")
            ignored_files.append(chroma_name)
            continue

        client_settings = Settings(
            is_persistent=True,
            persist_directory=chroma_name,
            anonymized_telemetry=False,
        )
        docsearch = Chroma(
            persist_directory=chroma_name, client_settings=client_settings
        )
        new_dbs = dbs(docsearch)

        if len(new_dbs.get_ids()) == 0:
            logging.warning(f"No vectors found in the chromadb directory {chroma_name}")
            print(f"No vectors found in the chromadb directory {chroma_name}")
            ignored_files.append(chroma_name)
            continue
        else:
            tot_db.merge(new_dbs)

        del docsearch, new_dbs
        gc.collect()

    return tot_db, ignored_files


def load_directory_db(
    directory_path: Union[str, Path],
    embeddings: Union[str, Embeddings, Callable],
    chunk_size: int,
    env_file: Optional[str] = "",
) -> dbs:
    ### get the chromadb directory name###
    if not isinstance(embeddings, str):
        embeddings, embeddings_name = handle_embeddings_and_name(
            embeddings, False, env_file
        )
    else:
        embeddings_name = embeddings
    embed_type, embed_name = separate_name(embeddings_name)

    storage_directory = get_storage_directory(
        directory_path, chunk_size, embed_type, embed_name
    )

    client_settings = Settings(
        is_persistent=True,
        persist_directory=storage_directory,
        anonymized_telemetry=False,
    )
    docsearch = Chroma(
        persist_directory=storage_directory, client_settings=client_settings
    )
    tot_dbs = dbs(docsearch)

    del docsearch
    gc.collect()

    if len(tot_dbs.get_ids()) == 0:
        raise ValueError(
            f"No vectors found in the chromadb directory {storage_directory}"
        )

    return tot_dbs


def load_files_db(
    file_path_list: Union[List[str], List[Path]],
    embeddings: Union[str, Embeddings, Callable],
    chunk_size: int,
    env_file: Optional[str] = "",
) -> dbs:
    dir_set = set()
    file_name_dict = defaultdict(list)
    tot_dbs = dbs()
    if not isinstance(embeddings, str):
        embeddings, embeddings_name = handle_embeddings_and_name(
            embeddings, False, env_file
        )
    else:
        embeddings_name = embeddings
    embed_type, embed_name = separate_name(embeddings_name)

    ### get the chromadb directory names and file names###
    for file_path in file_path_list:
        if not Path(file_path).exists():
            logging.warning(f"File {file_path} not found")
            print(f"File {file_path} not found")
            continue

        cur_file_name = Path(file_path).name
        cur_dir = Path(file_path).parent

        storage_directory = get_storage_directory(
            cur_dir, chunk_size, embed_type, embed_name
        )
        dir_set.add(storage_directory)
        file_name_dict[storage_directory].append(cur_file_name)

    for st_dir in dir_set:
        try:
            client_settings = Settings(
                is_persistent=True,
                persist_directory=st_dir,
                anonymized_telemetry=False,
            )
            docsearch = Chroma(
                persist_directory=st_dir, client_settings=client_settings
            )
            new_dbs = dbs(docsearch)
        except Exception as e:
            logging.warning(f"Error loading chromadb directory {st_dir}: {e}")
            print(f"Error loading chromadb directory {st_dir}: {e}")
            continue

        new_dbs = extract_db_by_file(new_dbs, file_name_dict[st_dir])
        tot_dbs.merge(new_dbs)

        del docsearch, new_dbs
        gc.collect()

    if len(tot_dbs.get_ids()) == 0:
        raise ValueError(
            f"No vectors found in the chromadb directory {storage_directory}"
        )

    return tot_dbs


def _display_db_num(dir_num: int, file_num: int, link_num: int, verbose: bool):
    if not verbose:
        return

    info = "\nload "
    if dir_num > 0:
        info += str(dir_num)
        if dir_num > 1:
            info += " directories, "
        else:
            info += " directory, "
    if file_num > 0:
        info += str(file_num)
        if file_num > 1:
            info += " files, "
        else:
            info += " file, "
    if link_num > 0:
        info += str(link_num)
        if link_num > 1:
            info += " links, "
        else:
            info += " link, "

    print(info + "\n\n")



from akasha.utils.atman import basic_llm
from akasha.utils.base import (
    DEFAULT_MODEL,
    DEFAULT_MAX_OUTPUT_TOKENS,
    DEFAULT_MAX_INPUT_TOKENS,
)
from akasha.utils.prompts.gen_prompt import (
    format_refine_summary_prompt,
    format_reduce_summary_prompt,
    format_sys_prompt,
)
from akasha.utils.prompts.format import (
    handle_params,
    handle_metrics,
    handle_table,
    language_dict,
)
from akasha.utils.db.load_docs import load_docs_from_info
from akasha.helper.base import get_doc_length, get_docs_length
from akasha.helper.run_llm import call_model
from typing import Union, Tuple
from pathlib import Path
from langchain.schema import Document
import time
import datetime
import math
import logging
from akasha.helper.token_counter import myTokenizer
from langchain.text_splitter import RecursiveCharacterTextSplitter
from tqdm import tqdm


class summary(basic_llm):
    def __init__(
        self,
        model: str = DEFAULT_MODEL,
        max_input_tokens: int = DEFAULT_MAX_INPUT_TOKENS,
        max_output_tokens: int = DEFAULT_MAX_OUTPUT_TOKENS,
        sum_type: str = "map_reduce",
        sum_len: int = 500,
        chunk_size: int = 500,
        chunk_overlap: int = 50,
        consecutive_merge_failures: int = 5,
        temperature: float = 0.0,
        prompt_format_type: str = "auto",
        language: str = "ch",
        record_exp: str = "",
        system_prompt: str = "",
        keep_logs: bool = False,
        verbose: bool = False,
        env_file: str = "",
    ):
        """ "initials of Summary class

        Args:
            **chunk_size (int, optional)**: chunk size of texts from documents. Defaults to 1000.\n
            **chunk_overlap (int, optional)**: chunk overlap of texts from documents. Defaults to 40.\n
            **model (str, optional)**: llm model to use. Defaults to "gpt-3.5-turbo".\n
            **verbose (bool, optional)**: show log texts or not. Defaults to False.\n
            **threshold (float, optional)**: (deprecated) the similarity threshold of searching. Defaults to 0.2.\n
            **language (str, optional)**: the language of documents and prompt, use to make sure docs won't exceed
                max token size of llm input.\n
            **record_exp (str, optional)**: use aiido to save running params and metrics to the remote mlflow or not if record_exp not empty, and set
                record_exp as experiment name.  default "".\n
            **system_prompt (str, optional)**: the system prompt that you assign special instruction to llm model, so will not be used
                in searching relevant documents. Defaults to "".\n
            **temperature (float, optional)**: temperature of llm model from 0.0 to 1.0 . Defaults to 0.0.\n
            **keep_logs (bool, optional)**: record logs or not. Defaults to False.\n
            **prompt_format_type (str, optional)**: the prompt and system prompt format for the language model, including auto, gpt, llama, chat_gpt, chat_mistral, chat_gemini . Defaults to "auto".
            **consecutive_merge_failures (int, optional)**: the number of consecutive merge failures before returning the current response list as the summary. Defaults to 5.
            **max_output_tokens (int, optional)**: max output tokens of llm model. Defaults to 1024.\n
            **max_input_tokens (int, optional)**: max input tokens of llm model. Defaults to 3000.\n
            **env_file (str, optional)**: the path of env file. Defaults to "".\n
        """
        super().__init__(
            model=model,
            max_input_tokens=max_input_tokens,
            max_output_tokens=max_output_tokens,
            temperature=temperature,
            language=language,
            record_exp=record_exp,
            system_prompt=system_prompt,
            keep_logs=keep_logs,
            verbose=verbose,
            env_file=env_file,
        )
        self.prompt_format_type = prompt_format_type
        self.consecutive_merge_failures = consecutive_merge_failures
        self.prompt = ""
        self.summary = ""
        self.docs = []
        self.prompt_tokens, self.prompt_length = 0, 0
        self.doc_tokens, self.doc_length = 0, 0
        self.sum_len = sum_len
        self.sum_type = sum_type
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap

    def _display_info(self, batch: int = 1) -> bool:
        if self.verbose is False:
            return False
        print(
            f"Model: {self.model}, Temperature: {self.temperature}, Summary Type: {self.sum_type}"
        )
        print(
            f"Prompt format type: {self.prompt_format_type}, Max input tokens: {self.max_input_tokens}"
        )
        print(
            f"Prompt tokens: {self.prompt_tokens}, Prompt length: {self.prompt_length}"
        )
        print(f"Doc tokens: {self.doc_tokens}, Doc length: {self.doc_length}")

        return True

    def _add_basic_log(
        self, timestamp: str, fn_type: str, history_messages: list = []
    ) -> bool:
        if super()._add_basic_log(timestamp, fn_type) is False:
            return False

        self.logs[timestamp]["sum_type"] = self.sum_type
        self.logs[timestamp]["sum_len"] = self.sum_len
        self.logs[timestamp]["consecutive_merge_failures"] = (
            self.consecutive_merge_failures
        )
        self.logs[timestamp]["chunk_size"] = self.chunk_size
        self.logs[timestamp]["chunk_overlap"] = self.chunk_overlap

        return True

    def _add_result_log(self, timestamp: str, time: float, reponse_list: list) -> bool:
        if super()._add_result_log(timestamp, time) is False:
            return False

        ### add token information ###
        self.logs[timestamp]["summary"] = self.summary
        self.logs[timestamp]["prompt_tokens"] = self.prompt_tokens
        self.logs[timestamp]["prompt_length"] = self.prompt_length
        self.logs[timestamp]["doc_tokens"] = self.doc_tokens
        self.logs[timestamp]["doc_length"] = self.doc_length
        self.logs[timestamp]["summary_list"] = reponse_list

        return True

    def _upload_logs(
        self, tot_time: float, doc_len: int, doc_tokens: int, response_list: list
    ) -> str:
        """_summary_

        Args:
            tot_time (float): _description_
            doc_len (int): _description_
            doc_tokens (int): _description_
        """
        if self.record_exp == "":
            return "no record_exp assigned, so no logs uploaded"

        params = handle_params(
            self.model,
            self.language,
            "",
            chunk_size=self.chunk_size,
        )
        params["chunk_overlap"] = self.chunk_overlap
        params["summary_type"] = "refine" if self.sum_type == "refine" else "map_reduce"
        metrics = handle_metrics(doc_len, tot_time, doc_tokens)
        table = handle_table(self.system_prompt, response_list, self.summary)
        from akasha.utils.upload import aiido_upload

        aiido_upload(self.record_exp, params, metrics, table)

        return "logs uploaded"

    def _reduce_summary(
        self,
        texts: list,
        tokens: int,
        total_list: list,
        progress: tqdm,
        pre_response_list_len: int,
        consecutive_merge_fail: int,
    ):
        """Summarize each chunk and merge them until the combined chunks are smaller than the maximum token limit.
        Then, generate the final summary. This method is faster and requires fewer tokens than the refine method.

        Args:
            **texts (list)**: list of texts from documents\n
            **tokens (int)**: used to save total tokens in recursive call.\n
            **total_list (list)**: used to save total response in recursive call.\n

        Returns:
            (list,int): llm response list and total tokens
        """
        response_list = []
        i = 0
        while i < len(texts):
            token, cur_text, newi = _get_text(
                texts, "", i, self.max_input_tokens, self.model
            )
            tokens += token

            progress.update(1)

            ### do the final summary if all chunks can be fits into llm model ###
            if i == 0 and newi == len(texts):
                prompt = format_reduce_summary_prompt(cur_text, self.sum_len)
                input_text = format_sys_prompt(
                    self.system_prompt,
                    "\n" + prompt,
                    self.prompt_format_type,
                    self.model,
                )
                response = call_model(
                    self.model_obj,
                    input_text,
                )

                total_list.append(response)

                return total_list, tokens

            prompt = format_reduce_summary_prompt(cur_text, 0)
            input_text = format_sys_prompt(
                self.system_prompt, "\n" + prompt, self.prompt_format_type, self.model
            )
            response = call_model(self.model_obj, input_text)

            i = newi

            response_list.append(response)
            total_list.append(response)

        ### handle merge fail ###
        if pre_response_list_len == len(response_list):
            consecutive_merge_fail += 1
            # if consecutive_merge_fail >= self.consecutive_merge_failures: return current response_list as summary #
            if consecutive_merge_fail >= self.consecutive_merge_failures:
                logging.warning(
                    "Cannot summarize due to texts too long, return current response_list as summary."
                )
                print(
                    "Cannot summarize due to texts too long, return current response_list as summary."
                )
                total_list.append("\n\n".join(response_list))
                return total_list, tokens
        else:
            consecutive_merge_fail = 0

        pre_response_list_len = len(response_list)

        return self._reduce_summary(
            response_list,
            tokens,
            total_list,
            progress,
            pre_response_list_len,
            consecutive_merge_fail,
        )

    def _refine_summary(self, texts: list) -> Union[list, int]:
        """refine summary summarizing a chunk at a time and using the previous summary as a prompt for
        summarizing the next chunk. This approach may be slower and require more tokens, but it results in a higher level of summary consistency.

        Args:
            **texts (list)**: list of texts from documents\n

        Returns:
            (list,int): llm response list and total tokens
        """
        ### setting variables ###
        previous_summary = ""
        i = 0
        tokens = 0
        response_list = []
        ###

        ### get tqdm progress bar and handle merge fail###
        progress = tqdm(total=len(texts), desc="Refine Summary")
        consecutive_merge_fail = 0
        fnsh_sum_list = []

        while i < len(texts):
            prei = i
            token, cur_text, i = _get_text(
                texts, previous_summary, i, self.max_input_tokens, self.model
            )
            tokens += token
            if previous_summary == "":
                prompt = format_reduce_summary_prompt(cur_text, self.sum_len)
            else:
                prompt = format_refine_summary_prompt(
                    cur_text, previous_summary, self.sum_len
                )
            text_input = format_sys_prompt(
                self.system_prompt, "\n" + prompt, self.prompt_format_type, self.model
            )
            response = call_model(
                self.model_obj,
                text_input,
            )

            response_list.append(response)
            previous_summary = response
            progress.update(i - prei)

            ### handle merge fail ###
            if prei == i:
                consecutive_merge_fail += 1
                if consecutive_merge_fail >= self.consecutive_merge_failures:
                    logging.warning(
                        "Cannot summarize current chunk due to texts too long, skip summarize current chunk."
                    )
                    fnsh_sum_list.append(response)
                    consecutive_merge_fail = 0
                    previous_summary = ""

            else:
                consecutive_merge_fail = 0

        progress.close()
        ## merge the failed summaries ##
        if len(fnsh_sum_list) > 0:
            response_list[-1] = "\n\n".join(fnsh_sum_list) + "\n\n" + response_list[-1]

        return response_list, tokens

    def __call__(self, content: Union[str, list, Path, Document], **kwargs) -> str:
        """input one or multiple content/source and return a summary of the content
        content can be a string, a list of strings, a path of file, a list of paths of files, a Document object, or a list of Document objects.
        Args:
            **content (Union[str, list, Path, Document])**:  the content you want to summarize, can be '.txt', '.docx', '.pdf' file.\n
            **sum_type (str, optional)**: summary method, "map_reduce" or "refine". Defaults to "map_reduce".\n
            **sum_len (int, optional)**: expected output length. Defaults to 500.\n
            **kwargs: the arguments you set in the initial of the class, you can change it here. Include:\n
                chunk_size, chunk_overlap, model, verbose, language , record_exp,
                system_prompt, max_input_tokens, temperature.
        Returns:
            str: the summary of the content.
        """

        self._set_model(**kwargs)
        self._change_variables(**kwargs)

        start_time = time.time()
        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        self.docs = load_docs_from_info(content)
        self.doc_length = get_docs_length(self.language, self.docs)
        self.doc_tokens = myTokenizer.compute_tokens(
            "".join([d.page_content for d in self.docs]), self.model
        )
        self.prompt_length = get_doc_length(self.language, self.system_prompt)
        self.prompt_tokens = myTokenizer.compute_tokens(self.system_prompt, self.model)

        ### check if docs do not has any content ###
        if self.doc_tokens == 0:
            raise ValueError("Can not get any text in the content")
        ### check if prompt <= max_input_tokens ###
        if self.prompt_tokens > self.max_input_tokens:
            raise ValueError("Prompt tokens exceed max_input_tokens")

        self._add_basic_log(timestamp, "summary")
        self._display_info()  # display the information of the parameters
        self._display_docs()

        text_splitter = RecursiveCharacterTextSplitter(
            separators=["\n", " ", ",", ".", "。", "!"],
            chunk_size=self.chunk_size,
            chunk_overlap=self.chunk_overlap,
        )
        split_docs = text_splitter.split_documents(self.docs)
        split_texts = [doc.page_content for doc in split_docs]

        ### start to ask llm ###
        if self.sum_type == "refine":
            response_list, self.doc_tokens = self._refine_summary(split_texts)

        else:
            per_sum_chunks = _calculate_per_summary_chunks(
                self.language, self.max_input_tokens, self.sum_len, self.chunk_size
            )
            approx_sum_times = _calculate_approx_sum_times(
                len(split_texts), per_sum_chunks
            )

            response_list_len = len(split_texts)
            consecutive_merge_fail = 0

            progress = tqdm(total=approx_sum_times, desc="Reduce_map Summary")
            response_list, self.doc_tokens = self._reduce_summary(
                split_texts, 0, [], progress, response_list_len, consecutive_merge_fail
            )
            progress.close()

        self.summary = response_list[-1]

        end_time = time.time()

        self._add_result_log(timestamp, end_time - start_time, response_list)

        self._upload_logs(
            end_time - start_time, self.doc_length, self.doc_tokens, response_list
        )

        return self.summary


def _calculate_approx_sum_times(chunks: int, per_sum_chunks: int) -> int:
    """approximate the total times of summarizing we need to do in reduce_summary method

    Args:
        chunks (int): number of chunks
        per_sum_chunks (int): number of chunks we can summarize in one time

    Returns:
        int: the total times of summarizing we need to do
    """
    times = 0
    while chunks > 1:
        chunks = math.ceil(chunks / per_sum_chunks)
        times += chunks
    return times


def _calculate_per_summary_chunks(
    language: str, max_input_tokens: int, summary_len: int, chunk_size: int
) -> int:
    """calculate the estimation of chunks that can fit into llm each time

    Args:
        language (str): texts language
        max_input_tokens (int): the max tokens we want to fit into llm model at one time
        summary_len (int): the length of summary tokens
        chunk_size (int): the chunk size of texts we want to summarize

    Returns:
        int: the estimation of chunks that can fit into llm each time
    """

    ret = 2

    if "chinese" in language_dict[language]:
        token_to_text = 2
    else:
        token_to_text = 1

    ret = max(ret, (token_to_text * max_input_tokens - summary_len) // chunk_size)

    return ret


def _get_text(
    texts: list,
    previous_summary: str,
    i: int,
    max_input_tokens: int,
    model_name: str = "openai:gpt-3.5-turbo",
) -> Tuple[int, str, int]:
    """used in summary, combine chunks of texts into one chunk that can fit into llm model

    Args:
        texts (list): chunks of texts
        previous_summary (str): _description_
        i (int): start from i-th chunk
        max_input_tokens (int): the max tokens we want to fit into llm model at one time
        model_name (str): model name(to calculate tokens) default "openai:gpt-3.5-turbo"\n

    Returns:
        (int, str, int): return the total tokens of combined chunks, combined chunks of texts, and the index of next chunk
    """
    cur_count = myTokenizer.compute_tokens(previous_summary, model_name)
    words_len = myTokenizer.compute_tokens(texts[i], model_name)
    cur_text = ""
    while cur_count + words_len < max_input_tokens and i < len(texts):
        cur_count += words_len
        cur_text += texts[i] + "\n"
        i += 1
        if i < len(texts):
            words_len = myTokenizer.compute_tokens(texts[i], model_name)

    return cur_count, cur_text, i
